/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid;

import java.time.LocalDate;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.lang3.StringUtils;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.stereotype.Component;

import de.ozgcloud.bescheid.binaryfile.BinaryFileService;
import de.ozgcloud.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.bescheid.nachricht.NachrichtService;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Component
@RequiredArgsConstructor
class BescheidEventListener {

	public static final String ORDER = "CREATE_BESCHEID";

	private static final String IS_CREATE_BESCHEID = "{T(de.ozgcloud.bescheid.BescheidEventListener).IS_CREATE_BESCHEID_COMMAND.test(event.getSource())}";

	public static final Predicate<Command> IS_CREATE_BESCHEID_COMMAND = command -> command.getOrder().equals(ORDER);

	static final String VORGANG_ID_BODYKEY = "vorgangId";
	static final String BESCHEID_VOM_BODYKEY = "bescheidVom";
	static final String GENEHMIGT_BODYKEY = "genehmigt";

	private static final String ERROR_MESSAGE = "Error on executing Create Bescheid Command.";

	private final BescheidService service;
	private final BinaryFileService fileService;
	private final NachrichtService nachrichtService;

	private final ApplicationEventPublisher eventPublisher;
	private final CurrentUserService userService;

	@EventListener(condition = IS_CREATE_BESCHEID)
	public void onCreateBescheidCommand(CommandCreatedEvent event) {
		Command command = event.getSource();

		doCreateBescheid(command);
	}

	private void doCreateBescheid(Command command) {
		SecurityContext prevContext = null;
		try {
			prevContext = userService.startSecurityContext(command);
			doCreateBescheidBiz(command);
			eventPublisher.publishEvent(new BescheidCreatedEvent(command));
		} catch (Exception e) {
			LOG.error("Error on executing Create Bescheid Command. Command failed.", e);
			eventPublisher.publishEvent(new CommandFailedEvent(command.getId(), buildErrorMessage(e)));
		} finally {
			userService.resetSecurityContext(prevContext);
		}
	}

	public void doCreateBescheidBiz(@NonNull Command command) {
		var bescheid = service.createBescheid(createRequest(command));
		bescheid = fileService.uploadBescheidFile(bescheid);

		nachrichtService.createNachrichtDraft(bescheid);
	}

	BescheidRequest createRequest(Command command) {
		var eventBody = command.getBodyObject();
		var builder = BescheidRequest.builder();

		builder.vorgangId(VorgangId.from(command.getVorgangId()));
		Optional.ofNullable(eventBody.get(BESCHEID_VOM_BODYKEY))
				.map(String.class::cast).map(LocalDate::parse).ifPresent(builder::bescheidVom);
		Optional.ofNullable(eventBody.get(GENEHMIGT_BODYKEY))
				.map(Object::toString).map(Boolean::valueOf)
				.ifPresentOrElse(builder::genehmigt, () -> builder.genehmigt(true));
		builder.createFor(userService.getUserProfile());

		return builder.build();
	}

	private String buildErrorMessage(Exception e) {
		try {
			StringBuilder sb = new StringBuilder(ERROR_MESSAGE);

			if (Objects.nonNull(e.getCause()) && StringUtils.isNotBlank(e.getCause().getMessage())) {
				sb.append(" Cause: " + e.getCause().getMessage());
			}

			if (e instanceof TechnicalException techException) {
				sb.append(" ExceptionId: " + techException);
			}

			return sb.toString();
		} catch (Exception e2) {
			LOG.error("Error in building Error Message (sick).", e2);
			return ERROR_MESSAGE;
		}
	}

}
