/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid;

import static de.ozgcloud.bescheid.BescheidEventListener.*;
import static de.ozgcloud.bescheid.BescheidRequestTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.security.core.context.SecurityContext;

import de.ozgcloud.bescheid.binaryfile.BinaryFileService;
import de.ozgcloud.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.bescheid.common.callcontext.UserProfile;
import de.ozgcloud.bescheid.common.callcontext.UserProfileTestFactory;
import de.ozgcloud.bescheid.nachricht.NachrichtService;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEventTestFactory;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandTestFactory;

class BescheidEventListenerTest {

	@Spy
	@InjectMocks
	private BescheidEventListener listener;

	@Mock
	private BescheidService service;
	@Mock
	private BinaryFileService fileService;
	@Mock
	private NachrichtService nachrichtService;

	@Mock
	private ApplicationEventPublisher eventPublisher;
	@Mock
	private CurrentUserService userService;

	@Nested
	class TestOnCreateBescheidCommand {

		private Command command = CommandTestFactory.createBuilder()
				.bodyObject(
						Map.of(VORGANG_ID_BODYKEY, VORGANG_ID.toString(),
								BESCHEID_VOM_BODYKEY, BESCHEID_VOM_STRING,
								GENEHMIGT_BODYKEY, GENEHMIGT))
				.build();

		@Mock
		private SecurityContext secContext;

		@Test
		void shouldCreateBescheid() {
			listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doCreateBescheidBiz(command);
		}

		@Test
		void shouldPublishErrorEventOnException() {
			doThrow(new RuntimeException("ups")).when(listener).doCreateBescheidBiz(any());

			listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(eventPublisher).publishEvent(any(CommandFailedEvent.class));
		}

		@Nested
		class HandleSecurityContext {

			@BeforeEach
			void init() {
				when(userService.startSecurityContext(any())).thenReturn(secContext);
			}

			@Test
			void shouldStartSecurityContext() {
				listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

				verify(userService).startSecurityContext(command);
			}

			@Test
			void shouldResetSecurityContext() {
				listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

				verify(userService).resetSecurityContext(secContext);
			}

			@Test
			void shouldResetSecurityContextAfterException() {
				doThrow(new RuntimeException("ups")).when(listener).doCreateBescheidBiz(any());

				listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

				verify(userService).resetSecurityContext(secContext);
			}
		}

		@Nested
		class CreateBescheidRequest {
			@Test
			void shouldContainUserProfile() {
				UserProfile user = UserProfileTestFactory.create();
				when(userService.getUserProfile()).thenReturn(user);

				var request = listener.createRequest(command);

				assertThat(request.getCreateFor()).isSameAs(user);
			}
		}

		@Nested
		class CreateBescheid {

			@Captor
			private ArgumentCaptor<BescheidRequest> requestCaptor;

			private Bescheid bescheid = BescheidTestFactory.create();
			private Bescheid bescheidWithFileId = BescheidTestFactory.create();

		@BeforeEach
		void init() {
			when(service.createBescheid(any())).thenReturn(bescheid);
			when(fileService.uploadBescheidFile(any())).thenReturn(bescheidWithFileId);
			when(userService.getUserProfile()).thenReturn(UserProfileTestFactory.create());
		}

			@Test
			void shouldCreateBescheid() {
				listener.doCreateBescheidBiz(command);

				verify(service).createBescheid(requestCaptor.capture());
				assertThat(requestCaptor.getValue()).usingRecursiveComparison()
						.isEqualTo(BescheidRequestTestFactory.createBuilder()
								.vorgangId(VorgangId.from(CommandTestFactory.VORGANG_ID))
								.build());
			}

			@Test
			void shouldSaveFile() {
				listener.doCreateBescheidBiz(command);

				verify(fileService).uploadBescheidFile(bescheid);
			}

			@Test
			void shouldSaveNachricht() {
				listener.doCreateBescheidBiz(command);

				verify(nachrichtService).createNachrichtDraft(bescheidWithFileId);
			}
		}
	}

}
