/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.smartdocuments;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.util.Optional;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.bescheid.BescheidRequestTestFactory;
import de.ozgcloud.bescheid.BescheidTestFactory;
import de.ozgcloud.bescheid.common.callcontext.UserProfileTestFactory;
import de.ozgcloud.bescheid.smartdocuments.SmartDocumentsRequest.CustomerData.UserData;
import de.ozgcloud.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.test.TestUtils;

class SmartDocumentsBescheidRemoteServiceTest {

	@Spy
	@InjectMocks
	private SmartDocumentsBescheidRemoteService service;

	@Mock
	private SmartDocumentsProperties properties;

	@Nested
	class TestBuildBescheid {

		@Test
		void shouldFillBescheid() {
			var bescheid = service.buildBescheid(BescheidRequestTestFactory.create(), SmartDocumentsResponseTestFactory.create());

			assertThat(bescheid).usingRecursiveComparison().ignoringFields("serviceKonto")
					.isEqualTo(BescheidTestFactory.createBuilder().nachrichtText(Optional.empty()).build());
		}
	}

	@Nested
	class TestCreateRequest {

		@Test
		void shouldFillUserData() {
			var request = service.createRequest(BescheidRequestTestFactory.create(), VorgangTestFactory.create());

			assertThat(request.getCustomerData().getUserData()).isNotNull()
					.extracting(UserData::getEmail).isEqualTo(UserProfileTestFactory.EMAIL);
		}
	}

	@Nested
	class TestGetNachrichtText {
		private File xmlFile = TempFileUtils.writeTmpFile(TestUtils.loadFile("SD_answer.xml"));

		@Test
		void shouldCallExtractText() {
			doReturn(Optional.of(xmlFile)).when(service).getXMLFile(any());

			service.getNachrichtText(SmartDocumentsResponseTestFactory.create());

			verify(service).extractTextFormXmlFile(notNull());
		}

		@Nested
		class ExtractingText {

			@Test
			void shouldReturnText() {
				var text = service.extractTextFormXmlFile(xmlFile);

				assertThat(text).isPresent().get().isNotNull();
			}

			@Test
			void shouldHandleError() {
				var text = service.extractTextFormXmlFile(null);

				assertThat(text).isEmpty();
			}

		}
	}
}
