/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.antragsraum;

import static java.util.Objects.*;

import jakarta.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.postfach.PostfachException;
import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;

@Service
@ConditionalOnProperty(AntragsraumProperties.PROPERTY_ANTRAGSRAUM_URL)
public class AntragsraumService {

	static final String USER_NOTIFICATION_TEMPLATE = """
			Guten Tag,
			  
			bei der Bearbeitung Ihres Antrags hat sich eine Rückfrage ergeben. Bitte beantworten Sie diese im Antragsraum:
			  
			%s
			  
			Vielen Dank,
			Ihre digitale Verwaltung
			""";

	@Autowired
	private AntragsraumProperties properties;

	@Autowired
	private NachrichtenManagerProperties nachrichtenManagerProperties;

	@Autowired
	private InfomanagerRemoteService infomanagerRemoteService;

	@PostConstruct
	void init() {
		if (isNull(nachrichtenManagerProperties.getUrl())) {
			throw new IllegalStateException("Address of Nachrichten-Manager is not set");
		}
	}

	public String getAntragsraumUrl() {
		return properties.getUrl();
	}

	public String getUserNotificationText() {
		return USER_NOTIFICATION_TEMPLATE.formatted(getAntragsraumUrl());
	}

	public void notifyAntragsraum(PostfachNachricht postfachNachricht) {
		try {
			infomanagerRemoteService.sendNotification(builInfomanagerNachricht(postfachNachricht));
		} catch (RuntimeException e) {
			throw new PostfachException("Error while sending notification to Antragsraum", PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE, e);
		}
	}

	InfomanagerNachricht builInfomanagerNachricht(PostfachNachricht postfachNachricht) {
		return InfomanagerNachricht.builder()
				.nachrichtId(postfachNachricht.getId())
				.vorgangId(postfachNachricht.getVorgangId())
				.postfachId(postfachNachricht.getPostfachId())
				.nachrichtenManagerUrl(nachrichtenManagerProperties.getUrl())
				.build();
	}
}
