/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.Base64;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.AttachmentFile;
import de.ozgcloud.nachrichten.postfach.BinaryFileService;
import de.ozgcloud.nachrichten.postfach.FileId;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import lombok.val;

@Service
public class MessageAttachmentService {

	@Autowired
	private PersistPostfachNachrichtService persistPostfachNachrichtService;

	@Autowired
	private BinaryFileService binaryFileService;

	public MessageAttachment getMessageAttachment(FileId fileId) {
		try {
			val metadata = binaryFileService.getFile(fileId).getMetadata();
			return MessageAttachment.builder()
					.fileName(metadata.getString("name"))
					.content(getAttachmentContent(fileId))
					.build();
		} catch (IOException e) {
			throw new TechnicalException(e.getMessage(), e);
		}
	}

	String getAttachmentContent(FileId fileId) throws IOException {
		ByteArrayOutputStream contentStream = new ByteArrayOutputStream();
		IOUtils.copy(getAttachmentContentStream(fileId), contentStream);
		return encodeAttachmentContent(contentStream.toByteArray());
	}

	private String encodeAttachmentContent(byte[] content) {
		return new String(Base64.getEncoder().encode(content));
	}

	public String persistAttachment(String vorgangId, MessageAttachment attachment) {
		return persistPostfachNachrichtService.persistAttachment(vorgangId, mapAttachmentFile(attachment));
	}

	AttachmentFile mapAttachmentFile(MessageAttachment attachment) {
		return AttachmentFile.builder()
				.name(attachment.getFileName())
				.content(() -> IOUtils.toInputStream(attachment.getContent(), Charset.defaultCharset())).build();
	}

	InputStream getAttachmentContentStream(FileId fileId) {
		return binaryFileService.getUploadedFileStream(fileId);
	}

}
