/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.antragsraum;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.postfach.PostfachException;
import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;

class AntragsraumServiceTest {

	@Spy
	@InjectMocks
	private AntragsraumService service;

	@Mock
	private AntragsraumProperties properties;
	@Mock
	private NachrichtenManagerProperties nachrichtenManagerProperties;
	@Mock
	private InfomanagerRemoteService infomanagerRemoteService;

	@Nested
	class TestGetAntragsraumUrl {

		private static final String URL = "url";

		@Test
		void shouldReturnUrl() {
			when(properties.getUrl()).thenReturn(URL);

			var url = service.getAntragsraumUrl();

			assertThat(url).isEqualTo(URL);
		}

	}

	@Nested
	class TestNotifyAntragsraum {

		private static final String NACHRICHTEN_MANAGER_URL = "nachrichtenManagerUrl";

		private PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@Test
		void shouldCallBuildInfomanagerNachricht() {
			service.notifyAntragsraum(postfachNachricht);

			verify(service).builInfomanagerNachricht(postfachNachricht);
		}

		@Test
		void shouldCallInfoManagerRemoteService() {
			var infomanagerNachricht = InfomanagerNachrichtTestFactory.create();
			doReturn(infomanagerNachricht).when(service).builInfomanagerNachricht(any());

			service.notifyAntragsraum(postfachNachricht);

			verify(infomanagerRemoteService).sendNotification(infomanagerNachricht);
		}

		@Test
		void shouldThrowPostfachException() {
			var causeException = new RuntimeException();
			when(infomanagerRemoteService.sendNotification(any())).thenThrow(causeException);

			var exception = assertThrows(PostfachException.class, () -> service.notifyAntragsraum(postfachNachricht));
			assertThat(exception.getMessageCode()).isEqualTo(PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE);
		}

		@Nested
		class TestBuildInfomanagerNachricht {

			@Test
			void shouldSetNachrichtId() {
				var infomanagerNachricht = buildInfomanagerNachricht();

				assertThat(infomanagerNachricht.getNachrichtId()).isEqualTo(PostfachNachrichtTestFactory.ID);
			}

			@Test
			void shouldSetVorgangId() {
				var infomanagerNachricht = buildInfomanagerNachricht();

				assertThat(infomanagerNachricht.getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetPostfachId() {
				var infomanagerNachricht = buildInfomanagerNachricht();

				assertThat(infomanagerNachricht.getPostfachId()).isEqualTo(MessageTestFactory.POSTFACH_ID);
			}

			@Test
			void shouldSetNachrichtenManagerUrl() {
				when(nachrichtenManagerProperties.getUrl()).thenReturn(NACHRICHTEN_MANAGER_URL);

				var infomanagerNachricht = buildInfomanagerNachricht();

				assertThat(infomanagerNachricht.getNachrichtenManagerUrl()).isEqualTo(NACHRICHTEN_MANAGER_URL);
			}

			private InfomanagerNachricht buildInfomanagerNachricht() {
				return service.builInfomanagerNachricht(postfachNachricht);
			}
		}
	}

}