/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Date;
import java.time.Instant;

import org.bson.BsonObjectId;
import org.bson.BsonValue;
import org.bson.Document;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import com.mongodb.client.gridfs.model.GridFSFile;

import de.ozgcloud.nachrichten.postfach.BinaryFileService;
import de.ozgcloud.nachrichten.postfach.FileId;

class MessageAttachmentServiceTest {
	private static final FileId FILE_ID = FileId.from("42");

	@InjectMocks
	private MessageAttachmentService messageAttachmentService;

	@Mock
	private BinaryFileService fileService;

	private GridFSFile gridFsfile;

	@Nested
	class TestLoadingAttachment {
		@BeforeEach
		void init() {
			Document metadata = new Document();
			metadata.put("name", MessageAttachmentTestFactory.FILENAME);
			BsonValue id = new BsonObjectId();
			gridFsfile = new GridFSFile(id, FILE_ID.toString(), 0, 0, Date.from(Instant.now()), metadata);

			when(fileService.getFile(any())).thenReturn(gridFsfile);
			when(fileService.getUploadedFileStream(any()))
					.thenReturn(new ByteArrayInputStream(MessageAttachmentTestFactory.DECODED_CONTENT.getBytes()));
		}

		@Test
		void shouldHaveAttachmentWithFileName() {
			MessageAttachment attachment = messageAttachmentService.getMessageAttachment(FILE_ID);

			assertThat(attachment.getFileName()).isEqualTo(MessageAttachmentTestFactory.FILENAME);
		}

		@Test
		void shouldHaveAttachmentContent() {
			MessageAttachment attachment = messageAttachmentService.getMessageAttachment(FILE_ID);

			assertThat(attachment.getContent()).isEqualTo(MessageAttachmentTestFactory.CONTENT);
		}

	}

	@Nested
	class TestLoadingAttachmentContent {
		@BeforeEach
		void init() {
			when(fileService.getUploadedFileStream(any()))
					.thenReturn(new ByteArrayInputStream(MessageAttachmentTestFactory.DECODED_CONTENT.getBytes()));
		}

		@Test
		void shouldGetInputStream() {
			InputStream input = messageAttachmentService.getAttachmentContentStream(FILE_ID);

			assertThat(input).isNotNull();
		}

		@Test
		void shouldGetContent() throws IOException {
			String input = messageAttachmentService.getAttachmentContent(FILE_ID);

			assertThat(input).isEqualTo(MessageAttachmentTestFactory.CONTENT);
		}
	}

	@Nested
	class TestMapAttachmentFile {

		@Test
		void shouldMapFileName() {
			var attachmentFile = messageAttachmentService.mapAttachmentFile(MessageAttachmentTestFactory.create());

			assertThat(attachmentFile.getName()).isEqualTo(MessageAttachmentTestFactory.FILENAME);
		}

		@Test
		void shouldMapContent() {
			var attachmentFile = messageAttachmentService.mapAttachmentFile(MessageAttachmentTestFactory.create());

			assertThat(attachmentFile.getContent()).hasContent(MessageAttachmentTestFactory.CONTENT);
		}
	}
}
