/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.List;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.nachrichten.postfach.FileId;
import de.ozgcloud.nachrichten.postfach.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;

class OsiPostfachMessageMapperTest {

	@InjectMocks
	private OsiPostfachMessageMapper mapper = Mappers.getMapper(OsiPostfachMessageMapper.class);
	@Mock
	private MessageAttachmentService messageAttachmentService;
	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@DisplayName("To mail")
	@Nested
	class TestToNachricht {

		@Test
		void shouldSetDirection() {
			var mail = mapper.toPostfachNachricht(MessageTestFactory.create());

			assertThat(mail.getDirection()).isEqualTo(Direction.IN);
		}

		@Test
		void shouldSetCreatedAt() {
			var mail = mapper.toPostfachNachricht(MessageTestFactory.create());

			assertThat(mail.getCreatedAt()).isNotNull().isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldIgnoreSentInformation() {
			var mail = mapper.toPostfachNachricht(MessageTestFactory.create());

			assertThat(mail.getSentAt()).isNull();
			assertThat(mail.getSentSuccessful()).isNull();
		}

		@Test
		void shouldPersistAttachment() {
			mapper.toPostfachNachricht(MessageTestFactory.create());

			verify(messageAttachmentService).persistAttachment(MessageTestFactory.VORGANG_ID, MessageTestFactory.ATTACHMENTS.get(0));
		}

		@Test
		void shouldSetPostfachAddress() {
			var expectedPostfachAddress = PostfachAddressTestFactory.createBuilder().serviceKontoType(OsiPostfachRemoteService.POSTFACH_TYPE).build();

			var mail = toMail();

			assertThat(mail.getPostfachAddress()).usingRecursiveComparison().isEqualTo(expectedPostfachAddress);
		}

		@Test
		void shouldPreserveNewLineInBody() {
			var bodyText = "line\nline";
			var mail = mapper.toPostfachNachricht(MessageTestFactory.createBuilder().mailBody(bodyText).build());

			assertThat(mail.getMailBody()).isEqualTo(bodyText);
		}

		@Test
		void shouldRecodeAND() {
			var mail = mapper.toPostfachNachricht(MessageTestFactory.createBuilder().mailBody("this&amp;that").build());

			assertThat(mail.getMailBody()).isEqualTo("this&that");
		}

		private PostfachNachricht toMail() {
			return mapper.toPostfachNachricht(MessageTestFactory.create());
		}
	}

	@DisplayName("To osi message")
	@Nested
	class TestToOsiMessage {

		@Test
		void shouldCallAttachmentService() {
			var fileId = FileId.from("42");

			toOsiMessage(PostfachNachrichtTestFactory.createBuilder().attachments(List.of(fileId.toString())).build());

			verify(messageAttachmentService).getMessageAttachment(fileId);
		}

		@Test
		void shouldMapFields() {
			var message = toOsiMessage(PostfachNachrichtTestFactory.create());

			assertThat(message.getSubject()).isEqualTo(MessageTestFactory.SUBJECT);
			assertThat(message.getAttachments()).hasSize(1);
		}

		@Test
		void shouldMapMailBody() {
			var message = toOsiMessage(PostfachNachrichtTestFactory.create());

			assertThat(message.getMailBody()).isEqualTo(MessageTestFactory.MAIL_BODY);
		}

		@Test
		void shouldReplaceNewLineToHtml() {
			var bodyText = "line\nline&";
			var message = toOsiMessage(PostfachNachrichtTestFactory.createBuilder().mailBody(bodyText).build());

			assertThat(message.getMailBody()).isEqualTo("line<br>line&");
		}

		@DisplayName("build postfachId")
		@Nested
		class TestToPostfachId {

			@DisplayName("by existing postfachAddress")
			@Nested
			class TestWithExistingPostfachAddress {

				@Test
				void shouldMapPostfachAddressToPostfachId() {
					var message = toOsiMessage(PostfachNachrichtTestFactory.create());

					assertThat(message.getPostfachId()).isEqualTo(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
				}
			}

			@DisplayName("by postfachId if postfachAddress not exists")
			@Nested
			class TestWithoutPostfachAddress {

				@Test
				void shouldMapPostfachAddressToPostfachId() {
					var message = toOsiMessage(PostfachNachrichtTestFactory.createBuilder().postfachAddress(null).build());

					assertThat(message.getPostfachId()).isEqualTo(MessageTestFactory.POSTFACH_ID);
				}
			}
		}

		private Message toOsiMessage(PostfachNachricht postfachNachricht) {
			return mapper.toOsiMessage(postfachNachricht);
		}
	}

	@DisplayName("To HTML body")
	@Nested
	class TestToHTMLBody {

		@Test
		void shouldNotContainNewlines() {
			var body = mapper.replaceNewlines(MessageTestFactory.MAIL_BODY);

			assertThat(body).doesNotContain("\n");
		}

	}
}