/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.antragsteller;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Objects;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.notification.command.CommandService;
import de.ozgcloud.notification.postfach.PostfachNachricht;
import de.ozgcloud.notification.vorgang.Vorgang;

@Service
class AntragstellerNotificationService {

	private static final String TEMPLATE_FILE = "antragsteller.notificationTemplate.txt";
	private static final String SUBJECT_TEMPLATE = "Eingangsbestätigung %s";

	private static final String TIMEZONE = "Europe/Berlin";
	private static final String DATETIME_FORMAT = "dd.MM.yyyy HH:mm:ss";

	@Autowired
	private CommandService commandService;

	public void sendNotification(Vorgang vorgang) {
		commandService.createSendPostfachNachrichtCommand(buildPostfachNachricht(vorgang));
	}

	PostfachNachricht buildPostfachNachricht(Vorgang vorgang) {
		return PostfachNachricht.builder()
				.vorgangId(vorgang.getId())
				.postfachId(vorgang.getPostfachId())
				.postfachAddress(vorgang.getPostfachAddress())
				.mailBody(formatMessage(vorgang))
				.subject(formatSubject(vorgang))
				.build();
	}

	String formatMessage(Vorgang vorgang) {
		return String.format(loadTemplate(),
				vorgang.getCreatedAt().withZoneSameInstant(ZoneId.of(TIMEZONE)).format(DateTimeFormatter.ofPattern(DATETIME_FORMAT)),
				vorgang.getVorgangNummer());
	}

	private String loadTemplate() {
		var templateInStream = AntragstellerNotificationService.class.getClassLoader().getResourceAsStream(TEMPLATE_FILE);
		if (Objects.isNull(templateInStream)) {
			throw new TechnicalException("Cannot find template file " + TEMPLATE_FILE);
		}
		try {
			return IOUtils.toString(templateInStream, StandardCharsets.UTF_8);
		} catch (IOException e) {
			throw new TechnicalException("Error reading template file", e);
		}
	}

	String formatSubject(Vorgang vorgang) {
		return String.format(SUBJECT_TEMPLATE, vorgang.getVorgangName());
	}
}
