/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.command;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.function.Predicate;

import org.apache.commons.beanutils.BeanMap;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.notification.command.Command.CommandOrder;
import de.ozgcloud.notification.postfach.PostfachAddress;
import de.ozgcloud.notification.postfach.PostfachAddressIdentifier;
import de.ozgcloud.notification.postfach.PostfachNachricht;
import de.ozgcloud.notification.postfach.StringBasedIdentifier;

@Service("notificationCommandService")
public class CommandService {

	private static final Predicate<Entry<Object, Object>> IS_NOT_CLASS_KEY = entry -> !StringUtils.equals(entry.getKey().toString(), "class");

	@Autowired
	private CommandRemoteService commandRemoteService;

	public void createSendPostfachNachrichtCommand(PostfachNachricht nachricht) {
		commandRemoteService.createCommand(buildCommand(nachricht));
	}

	Command buildCommand(PostfachNachricht nachricht) {
		return Command.builder()
				.vorgangId(nachricht.getVorgangId())
				.relationId(nachricht.getVorgangId().toString())
				.order(CommandOrder.SEND_POSTFACH_NACHRICHT)
				.body(toMap(nachricht))
				.build();
	}

	Map<String, Object> toMap(PostfachNachricht nachricht) {
		var result = new HashMap<String, Object>();

		new BeanMap(nachricht).entrySet().stream()
				.filter(entry -> Objects.nonNull(entry.getValue()))
				.filter(IS_NOT_CLASS_KEY)
				.forEach(entry -> result.put(entry.getKey().toString(), getValue(entry.getValue())));

		return Collections.unmodifiableMap(result);
	}
	
	private Object getValue(Object value) {
		if (value instanceof PostfachAddress postfachAddress) {
			return buildPostfachAddressMap(postfachAddress);
		}
		return value;
	}
	
	private Map<String, Object> buildPostfachAddressMap(PostfachAddress postfachAddress){
		 return Map.of(PostfachAddress.TYPE_FIELD, postfachAddress.getType(),
				 			PostfachAddress.VERSION_FIELD, postfachAddress.getVersion(), 
				 			PostfachAddress.IDENTIFIER_FIELD, buildPostfachAddressIdentifierMap(postfachAddress.getIdentifier()));
	}
	
	private Map<String, Object> buildPostfachAddressIdentifierMap(PostfachAddressIdentifier identifier){
		 return Map.of(StringBasedIdentifier.POSTFACH_ID_FIELD, ((StringBasedIdentifier) identifier).getPostfachId());
	}
}