/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.antragsteller;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.stream.IntStream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.notification.postfach.PostfachService;
import de.ozgcloud.notification.vorgang.Vorgang;
import de.ozgcloud.notification.vorgang.VorgangService;
import de.ozgcloud.notification.vorgang.VorgangTestFactory;
import lombok.SneakyThrows;

class AntragstellerNotificationEventListenerTest {

	@Spy
	@InjectMocks
	private AntragstellerNotificationEventListener eventListener;

	@Mock
	private AntragstellerNotificationService service;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private PostfachService postfachService;

	@DisplayName("On vorgang created")
	@Nested
	class TestOnVorgangCreated {

		private static final VorgangCreatedEvent EVENT = new VorgangCreatedEvent(VorgangTestFactory.ID.toString());
		private static final Vorgang VORGANG = VorgangTestFactory.createBuilder()
				.formEngineName(AntragstellerNotificationEventListener.FS_FORMENGINE_NAME).build();

		@Test
		void shouldLoadVorgang() {
			when(postfachService.isPostfachConfigured()).thenReturn(true);
			when(vorgangService.getVorgang(any())).thenReturn(VORGANG);

			onVorgangCreated();

			verify(vorgangService).getVorgang(VorgangTestFactory.ID);
		}

		@DisplayName("should not send notification when postfach is not configured")
		@Test
		void shouldNotSend() {
			when(postfachService.isPostfachConfigured()).thenReturn(false);

			onVorgangCreated();

			verify(vorgangService, never()).getVorgang(any());
		}

		@Nested
		class TestIsPostfachConfigured {

			@Test
			void shouldCallFetchIsPostfachService() {
				eventListener.isPostfachConfigured();

				verify(eventListener).fetchIsPostfachConfigured();
			}

			@Test
			void shouldNotCallFetchIsPostfachService() {
				ReflectionTestUtils.setField(eventListener, "isPostfachConfigured", Optional.of(true));

				eventListener.isPostfachConfigured();

				verify(eventListener, never()).fetchIsPostfachConfigured();
			}

			@Test
			void shouldReturnValue() {
				when(postfachService.isPostfachConfigured()).thenReturn(true);

				var isConfigured = eventListener.isPostfachConfigured();

				assertThat(isConfigured).isTrue();
			}

		}

		@Nested
		class TestFetchIsPostfachConfigured {

			@Test
			void shouldCallRemoteService() {
				eventListener.fetchIsPostfachConfigured();

				verify(postfachService).isPostfachConfigured();
			}

			@Test
			void shouldReturnValue() {
				when(postfachService.isPostfachConfigured()).thenReturn(true);

				var isConfigured = eventListener.fetchIsPostfachConfigured();

				assertThat(isConfigured).isTrue();
			}

			@Test
			void shouldCallIsPostfachConfiguredOnce() {
				fetchIsPostfachConfigured();

				verify(postfachService).isPostfachConfigured();
			}

			@SneakyThrows
			private void fetchIsPostfachConfigured() {
				var executorService = Executors.newFixedThreadPool(5);
				var countDownLatch = new CountDownLatch(3);

				IntStream.range(0, 3).forEach(i ->
						executorService.submit(() -> {
							eventListener.fetchIsPostfachConfigured();
							countDownLatch.countDown();
						})
				);

				countDownLatch.await();
			}
		}

		@DisplayName("send if required")
		@Nested
		class TestSendIfRequirest {

			@DisplayName("should not call service")
			@Nested
			class TestNotCallService {

				@BeforeEach
				void setup() {
					when(postfachService.isPostfachConfigured()).thenReturn(true);
				}

				@Test
				void onMissingPostfachIdAndPostfachAddress() {
					when(vorgangService.getVorgang(any())).thenReturn(VORGANG.toBuilder().postfachId(null).postfachAddress(null).build());

					onVorgangCreated();

					verify(service, never()).sendNotification(any());
				}

				@Test
				void forOtherThanFormSolutionsEngines() {
					when(vorgangService.getVorgang(any())).thenReturn(VorgangTestFactory.createBuilder().formEngineName("other").build());

					onVorgangCreated();

					verify(service, never()).sendNotification(any());
				}
			}

			@DisplayName("should call service")
			@Nested
			class TestCallService {

				@BeforeEach
				void setup() {
					when(postfachService.isPostfachConfigured()).thenReturn(true);
				}

				@Test
				void onExistingPostfachIdAndPostfachAddress() {
					when(vorgangService.getVorgang(any())).thenReturn(VORGANG);

					onVorgangCreated();

					verify(service).sendNotification(same(VORGANG));
				}

				@Test
				void onExistingPostfachIdAndMissingPostfachAddress() {
					when(vorgangService.getVorgang(any())).thenReturn(VORGANG.toBuilder().postfachAddress(null).build());

					onVorgangCreated();

					verify(service).sendNotification(any());
				}

				@Test
				void onMissingPostfachIdAndExistingPostfachAddress() {
					when(vorgangService.getVorgang(any())).thenReturn(VORGANG.toBuilder().postfachId(null).build());

					onVorgangCreated();

					verify(service).sendNotification(any());
				}
			}
		}

		private void onVorgangCreated() {
			eventListener.onVorgangCreated(EVENT);
		}
	}
}