/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.callcontext;

import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.AuthenticationTrustResolver;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

@Service
public class CurrentUserService {

	@Autowired
	private AuthenticationTrustResolver trustResolver;

	public CallContextUser getUser() {
		return findUser().orElseThrow(() -> new IllegalStateException("No authenticated User found"));
	}

	public Optional<CallContextUser> findUser() {
		return findTrustedAuthentication()
				.map(this::mapToCallContextUser)
				// remove this filter as soon we have a trusted authentication
				.filter(user -> StringUtils.isNotBlank(user.getClientName()));
	}

	private CallContextUser mapToCallContextUser(Authentication auth) {
		if (auth instanceof CallContextAuthenticationToken) {
			return (CallContextUser) auth.getPrincipal();
		} else {
			return CallContextUser.builder()
					.clientName(auth.getName())
					.organisatorischeEinheitenIds(
							auth.getAuthorities().stream().map(GrantedAuthority::getAuthority).collect(Collectors.toUnmodifiableSet()))
					.build();
		}
	}

	public Optional<Authentication> findAuthentication() {
		return findTrustedAuthentication();
	}

	Optional<Authentication> findTrustedAuthentication() {
		return Optional.ofNullable(SecurityContextHolder.getContext())
				.map(SecurityContext::getAuthentication)
				.filter(auth -> !trustResolver.isAnonymous(auth))
				.filter(Authentication::isAuthenticated);
	}
}