/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.callcontext;

import static de.ozgcloud.common.grpc.GrpcUtil.*;
import static de.ozgcloud.vorgang.callcontext.CallContextHandleInterceptor.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.security.core.context.SecurityContextHolder;

import de.ozgcloud.vorgang.callcontext.CallContextHandleInterceptor.LogContextSettingListener;
import io.grpc.Metadata;
import io.grpc.ServerCall;

class CallContextHandleInterceptorTest {

	@InjectMocks
	private CallContextHandleInterceptor interceptor;

	@Mock
	private ServerCall.Listener<Object> delegate;

	@Nested
	class TestCallListener {
		private LogContextSettingListener<?> listener;
		private Metadata headers = CallContextTestFactory.createMetadata();

		@BeforeEach
		void createListener() {
			listener = spy(interceptor.new LogContextSettingListener<>(delegate, headers));
		}

		@Nested
		class TestGetRequestId {
			@Test
			void shouldReturnIdFromHeader() {
				var reqId = listener.getRequestId();

				assertThat(reqId).isEqualTo(CallContextTestFactory.REQUEST_ID);
			}

			@Test
			void shouldReturnNewRequestIdIfNoGiven() {
				headers.removeAll(createKeyOf(KEY_REQUEST_ID));

				var reqId = listener.getRequestId();

				assertThat(reqId).isNotBlank().isNotEqualTo(CallContextTestFactory.REQUEST_ID);
			}

		}

		@DisplayName("Map headers to user")
		@Nested
		class TestHeadersToUser {
			@Test
			void shouldFilledUser() {
				var user = buildListener().createUser();

				assertThat(user).usingRecursiveComparison().isEqualTo(CallContextUserTestFactory.create());
			}

			@Test
			void shouldFillOrgaIdCollection() {
				Metadata metadata = CallContextTestFactory.createMetadata();
				metadata.put(createKeyOf(KEY_ACCESS_LIMITED_ORGAID), "orgaid_2".getBytes());

				var user = buildListener(metadata).createUser();

				assertThat(user.getOrganisatorischeEinheitenIds()).hasSize(2).contains(UserTestFactory.ORGANISATORISCHE_EINHEITEN_ID, "orgaid_2");
			}

			@Test
			void shouldFillEmptyListIfNoOrgaIds() {
				Metadata metadata = CallContextTestFactory.createMetadata();
				metadata.removeAll(createKeyOf(KEY_ACCESS_LIMITED_ORGAID));

				var user = buildListener(metadata).createUser();

				assertThat(user.getOrganisatorischeEinheitenIds()).isEmpty();
			}

			@DisplayName("for key isOrganisationEinheitenIdCheckNecessary")
			@Nested
			class TestIsOrganisationEinheitenIdCheckNecessary {

				@Test
				void shouldMapValueIfKeyIsPresent() {
					var metadata = CallContextTestFactory.createMetadata();

					var user = buildListener(metadata).createUser();

					assertThat(user.isOrganisationEinheitenIdCheckNecessary()).isTrue();
				}

				@Test
				void shouldMapFalseIfKeyIsNotPresent() {
					var metadata = CallContextTestFactory.createMetadata();
					metadata.removeAll(createKeyOf(KEY_ACCESS_LIMITED));

					var user = buildListener(metadata).createUser();

					assertThat(user.isOrganisationEinheitenIdCheckNecessary()).isFalse();
				}
			}

		}

		@Nested
		class TestStartSecurityContext {
			@Test
			void shouldHaveAuthenticatedAuthorization() {
				SecurityContextHolder.clearContext();

				listener.startSecurityContext();

				assertThat(SecurityContextHolder.getContext().getAuthentication().isAuthenticated()).isTrue();
			}

		}

		@Nested
		class TestDoSurround {

			private Runnable mockRunnable = mock(Runnable.class);

			@Test
			void shouldRunRunnable() {
				listener.doSurroundOn(mockRunnable);

				verify(mockRunnable).run();
			}

			@Test
			void shouldStartSecurityContext() {
				listener.doSurroundOn(mockRunnable);

				verify(listener).startSecurityContext();
			}

			@Test
			void shouldClearSecurityContext() {
				listener.doSurroundOn(mockRunnable);

				verify(listener).clearSecurityContext();
			}

			@Test
			void shouldClearSecurityContextOnException() {
				var testException = new RuntimeException() {
				};

				assertThatThrownBy(() -> listener.doSurroundOn(() -> {
					throw testException;
				})).isSameAs(testException);

				verify(listener).clearSecurityContext();
			}
		}

		@Nested
		class TestOnFunctions {

			private LogContextSettingListener<Object> listener;

			@BeforeEach
			void initListener() {
				listener = buildListener();
			}

			@Test
			void onMessageShouldCallSurround() {
				Object message = mock(Object.class);

				listener.onMessage(message);

				verify(listener).doSurroundOn(any());
			}

			@Test
			void onHalfCloseShouldCallSurround() {
				listener.onHalfClose();

				verify(listener).doSurroundOn(any());
			}

			@Test
			void onCancelShouldCallSurround() {
				listener.onCancel();

				verify(listener).doSurroundOn(any());
			}

			@Test
			void onCompleteShouldCallSurround() {
				listener.onComplete();

				verify(listener).doSurroundOn(any());
			}

			@Test
			void onReadyShouldCallSurround() {
				listener.onReady();

				verify(listener).doSurroundOn(any());
			}
		}

		private LogContextSettingListener<Object> buildListener() {
			return buildListener(headers);
		}

		private LogContextSettingListener<Object> buildListener(Metadata headers) {
			return spy(interceptor.new LogContextSettingListener<Object>(delegate, headers));
		}

	}

}
