/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import java.util.List;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Component;

import de.ozgcloud.vorgang.vorgang.Vorgang;

@Component
class CountStatisticMapper {

	static final Pattern PATH_COUNT_REQUEST_PATTERN = Pattern.compile(".+:.+");
	static final String PATH_COUNT_REQUEST_SEPARATOR = ":";

	GrpcVorgangCountResponse mapTo(Statistic statistic) {
		return GrpcVorgangCountResponse.newBuilder()
				.setByStatus(mapToByStatus(statistic))
				.addAllPathCountResult(mapToPathCountResults(statistic))
				.build();
	}

	GrpcByStatusResult mapToByStatus(Statistic statistic) {
		var counterByStatus = statistic.getCountByStatus();
		return GrpcByStatusResult.newBuilder()
				.setNeu(counterByStatus.get(Vorgang.Status.NEU))
				.setAngenommen(counterByStatus.get(Vorgang.Status.ANGENOMMEN))
				.setVerworfen(counterByStatus.get(Vorgang.Status.VERWORFEN))
				.setInBearbeitung(counterByStatus.get(Vorgang.Status.IN_BEARBEITUNG))
				.setBeschieden(counterByStatus.get(Vorgang.Status.BESCHIEDEN))
				.setAbgeschlossen(counterByStatus.get(Vorgang.Status.ABGESCHLOSSEN))
				.setWeitergeleitet(counterByStatus.get(Vorgang.Status.WEITERGELEITET))
				.build();
	}

	private List<GrpcPathCountResult> mapToPathCountResults(Statistic statistic) {
		return statistic.getCountResult().entrySet().stream().map(this::mapToPathCountResult).toList();
	}

	private GrpcPathCountResult mapToPathCountResult(Entry<String, Integer> countPathEntry) {
		return GrpcPathCountResult.newBuilder().setName(countPathEntry.getKey()).setValue(countPathEntry.getValue()).build();
	}

	public VorgangCountRequest mapFromVorgangCountRequest(GrpcVorgangCountRequest request) {
		return VorgangCountRequest.builder().countByPaths(mapFromCountByPathRequest(request.getCountByPathList())).build();
	}

	List<CountByPath> mapFromCountByPathRequest(List<String> countByPathRequests) {
		return countByPathRequests.stream().filter(StringUtils::isNoneBlank).map(this::mapPathCountRequest).toList();
	}

	CountByPath mapPathCountRequest(String pathCountRequest) {
		return CountByPath.builder().alias(extractAlias(pathCountRequest)).fieldPath(extractFieldPath(pathCountRequest)).build();
	}

	String extractFieldPath(String pathCountRequest) {
		validatePathCountRequest(pathCountRequest);
		return pathCountRequest.split(PATH_COUNT_REQUEST_SEPARATOR)[1];
	}

	String extractAlias(String pathCountRequest) {
		validatePathCountRequest(pathCountRequest);
		return pathCountRequest.split(PATH_COUNT_REQUEST_SEPARATOR)[0];
	}

	void validatePathCountRequest(String pathCountRequest) {
		if (!PATH_COUNT_REQUEST_PATTERN.matcher(pathCountRequest).matches()) {
			throw new VorgangStatisticBadRequestException(
					"Invalid path count request. Expected format 'alias:path.to.field' but was: '%s'".formatted(pathCountRequest));
		}
	}
}