/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.ozgcloud.vorgang.common.operator.OperandFunctionParser;
import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.vorgang.Vorgang;

@Component
class StatisticMapper {

	@Autowired
	private OperandFunctionParser operandFunctionParser;

	public List<VorgangStatisticRequest> mapFromVorgangStatisticRequest(GrpcVorgangStatisticRequest grpcRequest) {
		return grpcRequest.getQueryList().stream().map(this::mapFromVorgangStatisticQuery).toList();
	}

	VorgangStatisticRequest mapFromVorgangStatisticQuery(GrpcVorgangStatisticQuery statisticQuery) {
		return VorgangStatisticRequest.builder()
				.resultName(statisticQuery.getResultName())
				.fieldPath(getFieldPath(statisticQuery))
				.statisticGroupMethod(StatisticGroupMethod.from(statisticQuery.getGroupMethod()))
				.operator(buildOperator(statisticQuery))
				.build();
	}

	String getFieldPath(GrpcVorgangStatisticQuery statisticQuery) {
		return statisticQuery.getPath().startsWith(Vorgang.FIELD_CLIENT_ATTRIBUTES)
				? statisticQuery.getPath() + ".value"
				: statisticQuery.getPath();
	}

	Operator buildOperator(GrpcVorgangStatisticQuery statisticQuery) {
		return OperatorBuilder.from(statisticQuery.getOperator())
				.fieldPath(statisticQuery.getPath())
				.operand(getValue(statisticQuery)).build();
	}

	Object getValue(GrpcVorgangStatisticQuery statisticQuery) {
		return switch (statisticQuery.getOperandCase()) {
		case OPERANDBOOLVALUE -> statisticQuery.getOperandBoolValue();
		case OPERANDINTVALUE -> statisticQuery.getOperandIntValue();
		case OPERANDSTRINGVALUE -> getStringValue(statisticQuery);
		case OPERAND_NOT_SET -> null;
		};
	}

	String getStringValue(GrpcVorgangStatisticQuery statisticQuery) {
		var operandStringValue = statisticQuery.getOperandStringValue();
		if ("null".equalsIgnoreCase(operandStringValue)) {
			return null;
		} else if (operandStringValue.endsWith(")")) {
			return operandFunctionParser.parse(operandStringValue);
		}
		return operandStringValue;
	}

	public GrpcVorgangStatisticResponse mapToVorgangStatisticResponse(Statistic statistic) {
		return GrpcVorgangStatisticResponse.newBuilder()
				.addAllResult(mapCountByPathStatistic(statistic.getCountResult()))
				.addAllResult(mapIsFieldExistsStatistic(statistic.getExistsResult()))
				.build();
	}

	List<GrpcVorgangStatisticResult> mapCountByPathStatistic(Map<String, Integer> countByPathStatistic) {
		return countByPathStatistic.entrySet().stream().map(entry -> mapToCountResult(entry.getKey(), entry.getValue())).toList();
	}

	GrpcVorgangStatisticResult mapToCountResult(String key, Integer value) {
		return GrpcVorgangStatisticResult.newBuilder().setName(key).setResultIntValue(value).build();
	}

	List<GrpcVorgangStatisticResult> mapIsFieldExistsStatistic(Map<String, Boolean> isFieldExistsStatistic) {
		return isFieldExistsStatistic.entrySet().stream().map(entry -> mapToIsFieldExistsResult(entry.getKey(), entry.getValue())).toList();
	}

	GrpcVorgangStatisticResult mapToIsFieldExistsResult(String key, Boolean value) {
		return GrpcVorgangStatisticResult.newBuilder().setName(key).setResultBoolValue(value).build();
	}
}
