/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.status;

import java.util.ConcurrentModificationException;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
public class StatusService {
	private static final String CONCURRENT_MODIFICATION_ERROR_CODE = "concurrent_modification";

	@Autowired
	private ApplicationEventPublisher publisher;

	@Autowired
	private StatusRepository repository;

	void setStatusNeu(Command command) {
		executeStatusChangeCommand(command, Status.NEU);
	}

	public void annehmen(Command command) {
		executeStatusChangeCommand(command, Status.ANGENOMMEN);
	}

	public void verwerfen(Command command) {
		executeStatusChangeCommand(command, Status.VERWORFEN);
	}

	public void bescheiden(Command command) {
		executeStatusChangeCommand(command, Status.BESCHIEDEN);
	}

	public void abschliessen(Command command) {
		executeStatusChangeCommand(command, Status.ABGESCHLOSSEN);
	}

	public void wiedereroeffnen(Command command) {
		bearbeiten(command);
	}

	public void bearbeiten(Command command) {
		executeStatusChangeCommand(command, Status.IN_BEARBEITUNG);
	}

	public void zumLoeschenMarkieren(Command command) {
		executeStatusChangeCommand(command, Status.ZU_LOESCHEN);
	}

	public void executeStatusChangeCommand(Command command, Status status) {
		try {
			doUpdateStatus(command.getRelationId(), command.getRelationVersion(), status.name());
			publisher.publishEvent(new StatusChangedEvent(command));
		} catch (ConcurrentModificationException e) {
			LOG.warn("Concurrent modification on changeing status.");
			publisher.publishEvent(new CommandFailedEvent(command.getId(), CONCURRENT_MODIFICATION_ERROR_CODE));
		}
	}

	public void setStatusToInBearbeitung(String vorgangId, long version) {
		doUpdateStatus(vorgangId, version, Status.IN_BEARBEITUNG.name());
	}

	public void revokeStatusChange(Command command) {
		try {
			doUpdateStatus(command.getRelationId(), command.getRelationVersion() + 1,
					getPreviousStatus(command).orElseThrow(() -> new TechnicalException("Malformed or missing previous state for status revoke.")));

			publisher.publishEvent(new CommandRevokedEvent(command));
		} catch (ConcurrentModificationException e) {
			publisher.publishEvent(new CommandFailedEvent(command.getId(), CONCURRENT_MODIFICATION_ERROR_CODE));
		}
	}

	Optional<String> getPreviousStatus(Command command) {
		return Optional.ofNullable(((PersistedCommand) command).getPreviousState())
				.map(previousState -> MapUtils.getString(previousState, Vorgang.MONGODB_FIELDNAME_STATUS))
				.filter(Objects::nonNull);
	}

	void doUpdateStatus(String vorgangId, long version, String statusValue) {
		Map<String, Object> patch = Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, statusValue);

		repository.patch(vorgangId, version, patch);
	}

	public void setStatusToWeitergeleitet(String vorgangId, String commandId) {
		doUpdateStatusWithoutVerisonCheck(vorgangId, Status.WEITERGELEITET, commandId);
	}

	public void setStatusToInBearbeitung(String vorgangId, String commandId) {
		doUpdateStatusWithoutVerisonCheck(vorgangId, Status.IN_BEARBEITUNG, commandId);
	}

	void doUpdateStatusWithoutVerisonCheck(String vorgangId, Status newStatus, String commandId) {
		repository.updateStatusWithoutVersionCheck(vorgangId, newStatus);

		publisher.publishEvent(new StatusChangedEvent(commandId));
	}
}
