/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import java.util.ConcurrentModificationException;
import java.util.function.Predicate;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.nachrichten.email.MailSendErrorEvent;
import de.ozgcloud.nachrichten.email.MailSendRequest;
import de.ozgcloud.nachrichten.email.MailSentEvent;
import de.ozgcloud.vorgang.command.Order;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
public class ForwardingEventListener {

	private static final String IS_REDIRECT_EVENT_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.redirect.ForwardingEventListener).IS_REDIRECT_MAIL_REQ.test(event.getSource())}";
	private static final String IS_SUCCESSFULL_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.redirect.ForwardingEventListener).IS_SUCCESSFULL_ORDER.test(event.getSource())}";
	private static final String IS_FAILED_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.redirect.ForwardingEventListener).IS_FAILED_ORDER.test(event.getSource())}";
	private static final String IS_FORWARD_ORDER_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.redirect.ForwardingEventListener).IS_FORWARD_ORDER.test(event.getSource())}";

	public static final Predicate<MailSendRequest> IS_REDIRECT_MAIL_REQ = req -> req.getRequestReference() instanceof Forwarding;
	public static final Predicate<Command> IS_SUCCESSFULL_ORDER = command -> Order.FORWARD_SUCCESSFULL.isMeant(command.getOrder());
	public static final Predicate<Command> IS_FAILED_ORDER = command -> Order.FORWARD_FAILED.isMeant(command.getOrder());
	public static final Predicate<Command> IS_FORWARD_ORDER = command -> Order.REDIRECT_VORGANG.isMeant(command.getOrder());

	@Autowired
	private ForwardingService service;

	@Autowired
	private ApplicationEventPublisher publisher;

	@EventListener(condition = IS_FORWARD_ORDER_CONDITION)
	public void onForwardOrder(CommandCreatedEvent event) {
		handleException(() -> service.forwardByCommand(event.getSource()), event.getSource().getId());
	}

	@EventListener(condition = IS_REDIRECT_EVENT_CONDITION)
	public void triggerRedirectMailSent(MailSentEvent event) {
		var forwarding = (Forwarding) event.getSource().getRequestReference();

		service.onRedirectMailSent(forwarding);
	}

	@EventListener(condition = IS_REDIRECT_EVENT_CONDITION)
	public void publishCommandFailedEvent(MailSendErrorEvent event) {
		var forwarding = (Forwarding) event.getSource().getRequestReference();

		service.onRedirectMailFailed(forwarding, event.getErrorMessage());
	}

	@EventListener(condition = IS_SUCCESSFULL_ORDER_CONDITION)
	public void markAsSuccessfull(CommandCreatedEvent event) {
		handleException(() -> service.markAsSuccessfull(event.getSource()), event.getSource().getId());
	}

	@EventListener(condition = IS_FAILED_ORDER_CONDITION)
	public void markAsFailed(CommandCreatedEvent event) {
		service.markAsFailed(event.getSource());
	}

	private void handleException(Runnable runnable, String commandId) {
		try {
			runnable.run();
		} catch (ConcurrentModificationException e) {
			LOG.warn("Concurrent modification in executing Forwarding.");
			publishFailedEvent(commandId, "Concurrent modification in executing Forwarding.");
		} catch (RuntimeException e) {
			LOG.error("Command failed with Exception", e);
			publishFailedEvent(commandId, "Command Failed on Error");
		}
	}

	private void publishFailedEvent(String commandId, String message) {
		publisher.publishEvent(new CommandFailedEvent(commandId, message));
	}
}