/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.attributes;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.nachrichten.common.grpc.NachrichtenCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.Metadata;
import io.grpc.Metadata.Key;
import io.grpc.ServerCall;
import io.grpc.ServerCall.Listener;
import io.grpc.ServerCallHandler;
import io.grpc.ServerInterceptor;
import lombok.Getter;
import net.devh.boot.grpc.server.interceptor.GrpcGlobalServerInterceptor;

@SpringBootTest(classes = { VorgangManagerServerApplication.class }, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.pluto.address=in-process:test"
})
@DirtiesContext
@DataITCase
class ClientAttributeITCase {

	@Autowired
	private ClientAttributeService service;
	@Autowired
	private ServerTestCallContextHandleInterceptor serverTestInterceptor;

	// Um die Abhängigkeit zum VorgangManager-Code zu entfernen, kann man eine Server Dummy
	// Implementierung machen
	// siehe
	// https://yidongnan.github.io/grpc-spring-boot-starter/en/client/testing.html#running-a-dummy-server
	@MockBean
	private de.ozgcloud.vorgang.clientattribute.ClientAttributeService serverService;
	@Captor
	private ArgumentCaptor<ClientAttribute> attributeCaptor;

	@Nested
	class TestSetHasPostfachNachricht {

		@Test
		void shouldSetAttribute() {
			service.setHasPostfachNachricht(VorgangTestFactory.ID);

			verify(serverService, timeout(1000)).set(eq(VorgangTestFactory.ID), attributeCaptor.capture());
			var attribute = attributeCaptor.getValue();

			assertThat(attribute).usingRecursiveComparison().isEqualTo(createHasPostfachNachrichtAttribute());
		}

		@Test
		void shouldSetClient() {
			service.setHasPostfachNachricht(VorgangTestFactory.ID);

			verify(serverService, timeout(1000)).set(any(), any());
			var metadata = serverTestInterceptor.getLastHeaders();

			assertThat(metadata).isNotNull();
			assertThat(metadata.get(createKeyOf("CLIENT_NAME-bin"))).asString()
					.isEqualTo(NachrichtenCallContextAttachingInterceptor.NACHRICHTEN_MANAGER_CLIENT_NAME);
		}

		private ClientAttribute createHasPostfachNachrichtAttribute() {
			return de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory.createBuilder()
					.clientName(NachrichtenCallContextAttachingInterceptor.NACHRICHTEN_MANAGER_CLIENT_NAME)
					.attributeName(ClientAttributeService.ATTRIBUTE_NAME_HAS_POSTFACH_NACHRICHT)
					.boolValue(Optional.of(true)).intValue(Optional.empty())
					.build();
		}
	}

	// TODO move to a grpcUtil class in common
	static Key<byte[]> createKeyOf(String key) {
		return Key.of(key, Metadata.BINARY_BYTE_MARSHALLER);
	}
}

@Getter
@GrpcGlobalServerInterceptor
class ServerTestCallContextHandleInterceptor implements ServerInterceptor {

	private Metadata lastHeaders;

	@Override
	public <ReqT, RespT> Listener<ReqT> interceptCall(ServerCall<ReqT, RespT> call, Metadata headers, ServerCallHandler<ReqT, RespT> next) {
		lastHeaders = headers;

		return next.startCall(call, headers);
	}

}