/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.*;
import static org.springframework.test.web.client.response.MockRestResponseCreators.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.test.web.client.MockRestServiceServer;
import org.springframework.web.client.RestTemplate;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.nachrichten.postfach.FileId;
import de.ozgcloud.nachrichten.postfach.PostfachException;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;

@SpringBootTest(classes = { VorgangManagerServerApplication.class, OsiPostfachProperties.class }, properties = {
		"ozgcloud.osi.postfach.proxyapi.url=http://localhost/ApiProxy/V1/Message",
		"ozgcloud.osi.postfach.proxyapi.key=1234",
		"ozgcloud.osi.postfach.proxyapi.realm=test-realm",
		"ozgcloud.osi.postfach.scheduler.enabled=false"
})
@ITCase
class OsiPostfachRemoteServiceITCase {

	// properties, configured in application-itcase.yml
	static final String URL = "http://localhost/ApiProxy/V1/Message";
	static final String API_KEY = "1234";
	static final String API_REALM = "test-realm";

	@Autowired
	private OsiPostfachRemoteService service;

	@Autowired
	private RestTemplate restTemplate;

	@MockBean
	private MessageAttachmentService messageAttachmentService;

	private MockRestServiceServer mockServer;

	@BeforeEach
	void initMockServer() {
		mockServer = MockRestServiceServer.createServer(restTemplate);
		when(messageAttachmentService.getMessageAttachment(FileId.from(PostfachNachrichtTestFactory.ATTACHMENT_FILE_ID)))
				.thenReturn(MessageAttachmentTestFactory.create());
	}

	@Test
	void shouldCallWithApiKeyHeader() {
		mockServer.expect(header(PostfachConfiguration.HEADER_API_KEY, API_KEY)).andRespond(withSuccess());

		service.sendMessage(PostfachNachrichtTestFactory.create());

		mockServer.verify();
	}

	@Test
	void shouldCallWithRealmHeader() {
		mockServer.expect(header(PostfachConfiguration.HEADER_API_REALM, API_REALM)).andRespond(withSuccess());

		service.sendMessage(PostfachNachrichtTestFactory.create());

		mockServer.verify();
	}

	@Test
	void shouldRequestBaseUrl() {
		mockServer.expect(requestTo(URL)).andRespond(withSuccess());

		service.sendMessage(PostfachNachrichtTestFactory.create());

		mockServer.verify();
	}

	@Test
	void shouldSendHtmlBody() {
		mockServer.expect(jsonPath("$.isHtml").value(true))
				.andExpect(jsonPath("$.body").value(MessageTestFactory.MAIL_BODY))
				.andRespond(withSuccess());

		service.sendMessage(PostfachNachrichtTestFactory.create());

		mockServer.verify();
	}

	@Nested
	class TestSendMessage {

		@Test
		void shouldReturnSentSuccessful() {
			mockServer.expect(header(PostfachConfiguration.HEADER_API_REALM, API_REALM))
					.andRespond(withSuccess(String.valueOf(true), MediaType.APPLICATION_JSON));

			service.sendMessage(PostfachNachrichtTestFactory.create());

			mockServer.verify();
		}

		@Test
		void shouldReturnFalseOnError() {
			mockServer.expect(header(PostfachConfiguration.HEADER_API_REALM, API_REALM))
					.andRespond(withSuccess(String.valueOf(false), MediaType.APPLICATION_JSON));

			assertThrows(PostfachException.class, () -> service.sendMessage(PostfachNachrichtTestFactory.create()));// NOSONAR

			mockServer.verify();
		}
	}

	@Nested
	class TestGetMessages {

		@Test
		void shouldNotThrowErrorOnEmptyBody() {
			mockServer.expect(requestTo(URL)).andRespond(withSuccess());

			assertDoesNotThrow(() -> service.getAllMessages());

			mockServer.verify();
		}
	}
}