/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEventTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@ITCase
class VorgangAttachedItemEventListenerITCase {

	@Autowired
	private ApplicationEventPublisher publisher;

	@SpyBean
	private VorgangAttachedItemEventListener listener;
	@MockBean
	private VorgangAttachedItemService service;
	@MockBean
	private VorgangService vorgangService;
	@MockBean
	private CommandService commandService;

	@Nested
	class TestOnCreateItemCommand {
		@Test
		void shouldCallListener() {
			var command = CommandTestFactory.createBuilder().order(Order.CREATE_ATTACHED_ITEM.name()).relationId(VorgangTestFactory.ID)
					.bodyObject(VorgangAttachedItemTestFactory.asMap()).build();
			var event = CommandCreatedEventTestFactory.create(command);

			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(listener).createItem(any()));
		}
	}

	@Nested
	class TestOnUpdateItemCommand {
		@Test
		void shouldCallListener() {
			var command = CommandTestFactory.createBuilder().order(Order.UPDATE_ATTACHED_ITEM.name())
					.bodyObject(VorgangAttachedItemTestFactory.asMap()).build();
			var event = CommandCreatedEventTestFactory.create(command);

			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(listener).updateItem(any()));
		}
	}

	@Nested
	class TestOnPatchItemCommand {
		private Map<String, Object> bodyMap = Map.of("id", VorgangAttachedItemTestFactory.ID, "item", Map.of("done", true));

		@Test
		void shouldCallListener() {
			var command = CommandTestFactory.createBuilder().order(Order.PATCH_ATTACHED_ITEM.name())
					.bodyObject(bodyMap).build();
			var event = CommandCreatedEventTestFactory.create(command);

			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(listener).patchItem(any()));
		}
	}

	@Nested
	class TestOnDeleteItemCommand {

		@Test
		void shouldCallListener() {
			var command = CommandTestFactory.createBuilder().order(Order.DELETE_ATTACHED_ITEM.name()).build();
			var event = CommandCreatedEventTestFactory.create(command);

			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(listener).deleteItem(any()));
		}
	}

	@Nested
	class TestDeleteOnVorgangDeleted {

		@Test
		void shouldCallListener() {
			publisher.publishEvent(VorgangDeletedEventTestFactory.create());

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(listener).onVorgangDeleted(any()));
		}
	}

	@Nested
	class TestOnRevokeDeleteItem {

		@Test
		void shouldCallUnmarkAsDeleted() {
			var command = CommandTestFactory.createBuilder().order(Order.DELETE_ATTACHED_ITEM.name()).build();
			var event = new RevokeCommandEvent(command);

			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(
					() -> verify(service).unmarkAsDeleteByIdAndVersion(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION));
		}

	}

}
