/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.access.AccessDeniedException;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemResponse;
import io.grpc.stub.StreamObserver;

@DataITCase
class VorgangAttachedItemITCase {

	@Autowired
	private GrpcVorgangAttachedItemService service;

	@MockBean
	private CurrentUserService currentUserService;
	@Autowired
	private ApplicationEventPublisher publisher;
	@Autowired
	private MongoOperations mongoOperations;

	@BeforeEach
	void prepareDatabase() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
	}

	@DisplayName("Create item")
	@Nested
	class TestCreateItem {

		private Command command = CommandTestFactory.createBuilder().order(Order.CREATE_ATTACHED_ITEM.toString())
				.relationId(VorgangTestFactory.ID).bodyObject(VorgangAttachedItemTestFactory.asMap()).build();
		private CommandCreatedEvent event = CommandCreatedEventTestFactory.create(command);

		@Test
		void shouldPersistInDatabase() {
			publisher.publishEvent(event);
			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var loadedItems = mongoOperations.findAll(VorgangAttachedItem.class);

				assertThat(loadedItems).hasSize(1).first().satisfies(item -> {
					assertThat(item).usingRecursiveComparison().ignoringFields("id", "version").isEqualTo(VorgangAttachedItemTestFactory.create());
					assertThat(item.getId()).isNotNull();
					assertThat(item.getVersion()).isEqualTo(1L);
				});
			});
		}
	}

	@DisplayName("Update item")
	@Nested
	class TestUpdateItem {

		private Command command;
		private CommandCreatedEvent event;

		private VorgangAttachedItem persistedItem;

		@BeforeEach
		void step1prepareDatabase() {
			persistedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());

		}

		@BeforeEach
		void step2createEvent() {
			var itemMap = new HashMap<>(persistedItem.getItem());

			var bodyObjectMap = new HashMap<>(VorgangAttachedItemTestFactory.asMap());
			bodyObjectMap.put(VorgangAttachedItem.FIELDNAME_ITEM, itemMap);
			command = CommandTestFactory.createBuilder()
					.order(Order.UPDATE_ATTACHED_ITEM.name())
					.relationId(persistedItem.getId())
					.relationVersion(persistedItem.getVersion())
					.bodyObject(bodyObjectMap).build();

			event = CommandCreatedEventTestFactory.create(command);
		}

		@Test
		void shouldIncrementVersion() {
			publisher.publishEvent(event);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var loaded = mongoOperations.findAll(VorgangAttachedItem.class);

				assertThat(loaded).hasSize(1).first().satisfies(item -> {
					assertThat(item).usingRecursiveComparison().ignoringFields("id", "version").isEqualTo(VorgangAttachedItemTestFactory.create());
					assertThat(item.getVersion()).isEqualTo(2L);
				});
			});
		}
	}

	@DisplayName("Find")
	@Nested
	class TestFind {

		@Mock
		private StreamObserver<GrpcFindVorgangAttachedItemResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangAttachedItemResponse> responseCaptor;

		private Vorgang persistedVorgang;

		@DisplayName("with authorized User")
		@Nested
		class TestWithAuthorizedUser {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);

				persistedVorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build());
				mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0L).vorgangId(persistedVorgang.getId()).build());
			}

			@Test
			void shouldReturnResult() {
				var user = CallContextUserTestFactory.createBuilder()
						.organisationEinheitenIdCheckNecessary(true)
						.organisatorischeEinheitenId(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID)
						.build();
				when(currentUserService.findUser()).thenReturn(Optional.of(user));

				assertDoesNotThrow(() -> doFind());
			}

			// TODO remove this Test. Implement PoicyServiceITCase to test permission check
			// and verify service call in Unit Test
			@Test
			void shouldThrowAccessDeniedException() {
				var user = CallContextUserTestFactory.createBuilder()
						.organisationEinheitenIdCheckNecessary(true)
						.clearOrganisatorischeEinheitenIds().organisatorischeEinheitenId("doesNotMatch")
						.build();
				when(currentUserService.findUser()).thenReturn(Optional.of(user));

				assertThatThrownBy(() -> doFind()).isInstanceOf(AccessDeniedException.class);
			}
		}

		private GrpcFindVorgangAttachedItemResponse doFind() {
			service.find(GrpcFindVorgangAttachedItemRequestTestFactory.createBuilder().setVorgangId(persistedVorgang.getId()).build(),
					responseObserver);
			verify(responseObserver).onNext(responseCaptor.capture());

			return responseCaptor.getValue();
		}
	}

	@DisplayName("by Id")
	@Nested
	class TestGetById {

		@Mock
		private StreamObserver<GrpcVorgangAttachedItemResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcVorgangAttachedItemResponse> responseCaptor;

		private GrpcVorgangAttachedItemRequest request;
		private VorgangAttachedItem persistedItem;

		@Nested
		class TestWithAuthorizedUser {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);

				var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
				persistedItem = mongoOperations
						.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0L).vorgangId(vorgang.getId()).build());

				request = GrpcVorgangAttachedItemRequestTestFactory.createBuilder().setId(persistedItem.getId()).build();
			}

			@Test
			void shouldReturnResult() {
				var user = CallContextUserTestFactory.createBuilder()
						.organisatorischeEinheitenId(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID)
						.build();
				when(currentUserService.findUser()).thenReturn(Optional.of(user));

				assertDoesNotThrow(() -> doGetById());
			}

			@Test
			void shouldThrowAccessDeniedException() {
				var user = CallContextUserTestFactory.createBuilder()
						.clearOrganisatorischeEinheitenIds().organisatorischeEinheitenId("doesNotMatch")
						.build();
				when(currentUserService.findUser()).thenReturn(Optional.of(user));

				assertThatThrownBy(() -> doGetById()).isInstanceOf(AccessDeniedException.class);
			}
		}

		private GrpcVorgangAttachedItemResponse doGetById() {
			service.getById(request, responseObserver);
			verify(responseObserver).onNext(responseCaptor.capture());

			return responseCaptor.getValue();
		}
	}
}