/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;

import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class VorgangAttachedItemRepositoryITCase {

	@Autowired
	private VorgangAttachedItemRepository repository;
	@Autowired
	private MongoOperations mongoOperations;

	private VorgangAttachedItem persistedItem;

	@BeforeEach
	void prepareDatabase() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
	}

	@DisplayName("Find by")
	@Nested
	class TestFindBy {

		@BeforeEach
		void init() {
			persistedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
		}

		@Nested
		class VorgangIdAndClientAndItemName {

			@Test
			void shouldReturnResult() {
				var loadedItems = repository.findByVorgangIdAndClientAndItemNameAndDeleted(VorgangTestFactory.ID,
						VorgangAttachedItemTestFactory.CLIENT,
						VorgangAttachedItemTestFactory.ITEM_NAME, VorgangAttachedItemService.NOT_DELETED);

				assertThat(loadedItems).hasSize(1);
			}

			@Test
			void shouldReturnEmptyResult() {
				var loadedItems = repository.findByVorgangIdAndClientAndItemNameAndDeleted(VorgangTestFactory.ID,
						VorgangAttachedItemTestFactory.CLIENT,
						VorgangAttachedItemTestFactory.ITEM_NAME, VorgangAttachedItemTestFactory.DELETED);

				assertThat(loadedItems).isEmpty();
			}
		}

		@Nested
		class VorgangIdAndClient {

			@Test
			void shouldReturnResult() {
				var loadedItems = repository.findByVorgangIdAndClientAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemTestFactory.CLIENT,
						VorgangAttachedItemService.NOT_DELETED);

				assertThat(loadedItems).hasSize(1);
			}

			@Test
			void shouldReturnEmptyResult() {
				var loadedItems = repository.findByVorgangIdAndClientAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemTestFactory.CLIENT,
						VorgangAttachedItemTestFactory.DELETED);

				assertThat(loadedItems).isEmpty();
			}
		}

		@Nested
		class VorgangIdAndItemName {

			@Test
			void shouldReturnResult() {
				var loadedItems = repository.findByVorgangIdAndItemNameAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemTestFactory.ITEM_NAME,
						VorgangAttachedItemService.NOT_DELETED);

				assertThat(loadedItems).hasSize(1);
			}

			@Test
			void shouldReturnEmptyResult() {
				var loadedItems = repository.findByVorgangIdAndItemNameAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemTestFactory.ITEM_NAME,
						VorgangAttachedItemTestFactory.DELETED);

				assertThat(loadedItems).isEmpty();
			}
		}

		@Nested
		class VorgangId {

			@Test
			void shouldReturnResult() {
				var loadedItems = repository.findByVorgangIdAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);

				assertThat(loadedItems).hasSize(1);
			}

			@Test
			void shouldReturnEmptyResult() {
				var loadedItems = repository.findByVorgangIdAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemTestFactory.DELETED);

				assertThat(loadedItems).isEmpty();
			}
		}
	}

	@Nested
	class TestPatch {

		@BeforeEach
		void prepareDatabase() {
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

			persistedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
		}

		@Test
		void shouldUpdateFieldValue() {
			repository.patch(persistedItem.getId(), persistedItem.getVersion(), buildPatchMap());

			var loaded = findVorgangAttachedItem();
			assertThat(loaded).isNotNull();
			assertThat(loaded.getItem()).containsOnly(
					entry("dontTouch", 600),
					entry(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME, VorgangAttachedItemTestFactory.ITEM_FIELD_STRING_VALUE));
		}

		@Test
		void shouldUpdateWithNullValue() {
			Map<String, Object> sourceMap = Collections.singletonMap(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME, null);

			repository.patch(persistedItem.getId(), persistedItem.getVersion(), sourceMap);

			var loaded = findVorgangAttachedItem();
			assertThat(loaded.getItem()).containsOnly(entry(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME, null));
		}

		@Test
		void shouldIncreaseVersion() {
			repository.patch(persistedItem.getId(), persistedItem.getVersion(), buildPatchMap());

			var loaded = findVorgangAttachedItem();
			assertThat(loaded).isNotNull();
			assertThat(loaded.getVersion()).isEqualTo(persistedItem.getVersion() + 1);
		}

		@Test
		void shouldNotUpdateOnWrongVersion() {
			repository.patch(persistedItem.getId(), 42, buildPatchMap());

			var loaded = findVorgangAttachedItem();

			assertThat(loaded).usingRecursiveComparison().isEqualTo(persistedItem);
		}

		@Test
		void shouldNotPatchDeletedItem() {
			var deletedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).deleted(true).build());

			repository.patch(deletedItem.getId(), deletedItem.getVersion(), buildPatchMap());

			var loaded = mongoOperations.findById(deletedItem.getId(), VorgangAttachedItem.class);
			assertThat(loaded).usingRecursiveComparison().isEqualTo(deletedItem);
		}

		private Map<String, Object> buildPatchMap() {
			return Map.of("dontTouch", 600);
		}
	}

	@Nested
	class TestForceUpdate {

		@Test
		void shouldNotPatchDeletedItem() {
			var deletedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).deleted(true).build());

			repository.forcePatch(deletedItem.getId(), buildPatchMap());

			var loaded = mongoOperations.findById(deletedItem.getId(), VorgangAttachedItem.class);
			assertThat(loaded).usingRecursiveComparison().isEqualTo(deletedItem);
		}

		private Map<String, Object> buildPatchMap() {
			return Map.of("dontTouch", 600);
		}
	}

	@Nested
	class TestDelete {

		@BeforeEach
		void setup() {
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
			persistedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
			mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
		}

		@Test
		void shouldDeleteByVorgangId() {
			repository.deleteByVorgangId(VorgangTestFactory.ID);

			assertThat(mongoOperations.findAll(VorgangAttachedItem.class)).isEmpty();
		}

		@Test
		void shouldDeleteByIdAndVersion() {
			repository.delete(persistedItem.getId(), persistedItem.getVersion());

			assertThat(repository.findById(persistedItem.getId())).isEmpty();
		}

		@Test
		void shouldThrowExceptionWhenVersionMismatch() {
			var expectedId = persistedItem.getId();
			var expectedVersion = persistedItem.getVersion();
			mongoOperations.save(persistedItem);

			assertThrows(ConcurrentModificationException.class, () -> repository.delete(expectedId, expectedVersion));
		}
	}

	@Nested
	class TestMarkAsDeleted {

		private static final boolean IS_DELETED = true;

		@BeforeEach
		void setup() {
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
			persistedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
			mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0).build());
		}

		@Test
		void shouldMarkAsDeleted() {
			repository.updateDeleted(IS_DELETED, persistedItem.getId(), persistedItem.getVersion());

			var loaded = findVorgangAttachedItem();
			assertThat(loaded.isDeleted()).isTrue();
		}

		@Test
		void shouldThrowExceptionWhenVersionMismatch() {
			var expectedId = persistedItem.getId();
			var expectedVersion = persistedItem.getVersion();
			mongoOperations.save(persistedItem);

			assertThrows(ConcurrentModificationException.class, () -> repository.delete(expectedId, expectedVersion));
		}
	}

	private VorgangAttachedItem findVorgangAttachedItem() {
		return mongoOperations.findById(persistedItem.getId(), VorgangAttachedItem.class);
	}
}