/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute.AccessPermission;

class ClientAttributeReadPermittedTest {

	@InjectMocks
	private ClientAttributeReadPermitted predicate;

	@Mock
	private CurrentUserService userService;

	@BeforeEach
	void mockUserService() {
		when(userService.getUser()).thenReturn(CallContextUserTestFactory.create());
	}

	@Test
	void shouldReturnTrueForSameClient() {
		var result = predicate.test(ClientAttributeTestFactory.create());

		assertThat(result).isTrue();
	}

	@Nested
	class TestPrivateAccess {

		@Test
		void shouldReturnTrueForSameClient() {
			var result = predicate.test(ClientAttributeTestFactory.createBuilder().access(AccessPermission.PRIVATE).build());

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseForOtherClient() {
			when(userService.getUser()).thenReturn(CallContextUserTestFactory.createBuilder().clientName("other").build());

			var result = predicate.test(ClientAttributeTestFactory.createBuilder().access(AccessPermission.PRIVATE).build());

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestReadOnlyAndReadWriteAccess {

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = "PRIVATE")
		void shouldReturnTrueForSameClient(AccessPermission access) {
			var result = predicate.test(ClientAttributeTestFactory.createBuilder().access(access).build());

			assertThat(result).isTrue();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = "PRIVATE")
		void shouldReturnTrueForOtherClient(AccessPermission access) {
			when(userService.getUser()).thenReturn(CallContextUserTestFactory.createBuilder().clientName("other").build());

			var result = predicate.test(ClientAttributeTestFactory.createBuilder().access(access).build());

			assertThat(result).isTrue();
		}
	}

}
