/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.assertj.core.api.InstanceOfAssertFactories;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class CommandRepositoryITCase {

	@Autowired
	private CommandRepository repository;
	@Autowired
	private MongoOperations mongoOperations;

	@BeforeEach
	void clearDB() {
		mongoOperations.dropCollection(Command.class);
	}

	@Nested
	class TestSave {

		@Test
		void shouldSave() {
			repository.save(CommandTestFactory.create());

			verifyCommandExists(CommandTestFactory.ID);
		}

		@Test
		void shouldGenerateIdOnSave() {
			Command command = repository.save(CommandTestFactory.createBuilder().id(null).build());

			verifyCommandExists(command.getId());
		}

		private void verifyCommandExists(String id) {
			Command command = mongoOperations.findById(id, Command.class);

			assertThat(command).isNotNull();
			assertThat(command.getId()).isEqualTo(id);
		}
	}

	@Nested
	class TestFinishCommand {

		@Test
		void shouldUpdateStatusToFinished() {
			var command = mongoOperations.save(CommandTestFactory.create());

			repository.finishCommand(command.getId());

			var result = repository.getById(command.getId()).get();
			assertThat(result.getStatus()).isEqualTo(CommandStatus.FINISHED);
		}

		@Test
		void shouldSetFinishedAt() {
			var command = mongoOperations.save(CommandTestFactory.createBuilder().createdAt(null).build());

			repository.finishCommand(command.getId());

			var result = repository.getById(command.getId()).get();
			assertThat(result.getFinishedAt()).isNotNull();
		}

		@Test
		void shouldUpdateStatusToFinishedWithCreatedResource() {
			var command = mongoOperations.save(CommandTestFactory.create());

			repository.finishCommand(command.getId(), CommandTestFactory.CREATED_RESOURCE);

			var result = repository.getById(command.getId()).get();
			assertThat(result.getStatus()).isEqualTo(CommandStatus.FINISHED);
		}

		@Test
		void shouldSetFinishedAtWithCreatedResource() {
			var command = mongoOperations.save(CommandTestFactory.createBuilder().createdAt(null).build());

			repository.finishCommand(command.getId(), CommandTestFactory.CREATED_RESOURCE);

			var result = repository.getById(command.getId()).get();
			assertThat(result.getFinishedAt()).isNotNull();
		}

		@Test
		void shouldSetCreatedResource() {
			var command = CommandTestFactory.createBuilder().id(null).build();
			var savedCommand = repository.save(command);
			assertThat(savedCommand.getCreatedResource()).isNull();

			repository.finishCommand(savedCommand.getId(), CommandTestFactory.CREATED_RESOURCE);

			var result = repository.getById(savedCommand.getId()).get();
			assertThat(result.getCreatedResource()).isEqualTo(CommandTestFactory.CREATED_RESOURCE);
		}

		@Test
		void shouldNotOverrideCreatedResource() {
			var command = CommandTestFactory.createBuilder().id(null).createdResource(CommandTestFactory.CREATED_RESOURCE).build();
			var savedCommand = repository.save(command);

			repository.finishCommand(savedCommand.getId(), CommandTestFactory.CREATED_RESOURCE);

			var result = repository.getById(savedCommand.getId()).get();
			assertThat(result.getCreatedResource()).isEqualTo(CommandTestFactory.CREATED_RESOURCE);
		}
	}

	@Nested
	class TestGetById {

		private final String ADDITIONAL_COMMAND_ID = UUID.randomUUID().toString();

		@BeforeEach
		void persistAdditionalCommandToVorgang() {
			mongoOperations.save(CommandTestFactory.create());
		}

		@Test
		void shouldFindCommand() {
			verifyCommandExists(CommandTestFactory.ID);
		}

		@Test
		void shouldFindAdditionalCommand() {
			mongoOperations.save(CommandTestFactory.createBuilder().id(ADDITIONAL_COMMAND_ID).build());

			verifyCommandExists(ADDITIONAL_COMMAND_ID);
		}

		private void verifyCommandExists(String id) {
			Optional<Command> command = repository.getById(id);

			assertThat(command).isPresent();
			assertThat(command.get().getId()).isEqualTo(id);
		}
	}

	@Nested
	class TestUpdateCommandStatus {

		@BeforeEach
		void persistVorgangWithCommand() {
			repository.save(CommandTestFactory.create());
		}

		@ParameterizedTest
		@EnumSource(CommandStatus.class)
		void shouldUpdateCommandStatus(CommandStatus status) {
			repository.updateCommandStatus(CommandTestFactory.ID, status);

			Optional<Command> command = repository.getById(CommandTestFactory.ID);

			assertThat(command).isPresent();
			assertThat(command.get().getStatus()).isEqualTo(status);
		}
	}

	@Nested
	class TestUpdateCommandStatusAndVersion {

		@BeforeEach
		void persistVorgangWithCommand() {
			repository.save(CommandTestFactory.create());
		}

		@ParameterizedTest
		@EnumSource(CommandStatus.class)
		void shouldUpdateCommandStatus(CommandStatus status) {
			repository.updateCommandStatusAndVersion(CommandTestFactory.ID, status, 78L);

			Optional<Command> command = repository.getById(CommandTestFactory.ID);

			assertThat(command).isPresent();
			assertThat(command.get().getRelationVersion()).isEqualTo(78L);
			assertThat(command.get().getStatus()).isEqualTo(status);
		}
	}

	@Nested
	class TestGetPendingCommands {

		@BeforeEach
		void persistCommand() {
			repository.save(CommandTestFactory.create());
			repository.save(CommandTestFactory.createBuilder().id(null).vorgangId(VorgangTestFactory.ID).status(CommandStatus.FINISHED).build());
		}

		@Test
		void shouldReturnPendingCommand() {
			var result = repository.getPendingCommands(VorgangTestFactory.ID);

			assertThat(result).hasSize(1);
			assertThat(result.get(0).getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
			assertThat(result.get(0).getStatus()).isEqualTo(CommandStatus.PENDING);
		}
	}

	@Nested
	class TestFindCommands {

		@Test
		void shouldReturnCommandByVorgang() {
			repository.save(CommandTestFactory.create());

			var commands = repository.findCommands(VorgangTestFactory.ID, Collections.emptyList(), Optional.empty());

			assertThat(commands).hasSize(1);
		}

		@Nested
		@DisplayName("by order, status and createdBy")
		class TestByOrderStatusTime {

			@BeforeEach
			void setup() {
				repository.save(createCommandBuilder().build());
				repository.save(createCommandBuilder().status(CommandStatus.FINISHED).build());
				repository.save(createCommandBuilder().status(CommandStatus.FINISHED).order(Order.VORGANG_LOESCHEN.name()).build());
			}

			private PersistedCommand.PersistedCommandBuilder createCommandBuilder() {
				return CommandTestFactory.createBuilder().id(null).createdAt(getNow());
			}

			private ZonedDateTime getNow() {
				return ZonedDateTime.now(ZoneOffset.UTC).withNano(0);
			}

			@Test
				// TODO schreibe ein Testcase für Order, Status und Zeit
			void shouldFindCommand() {
				var command = repository.save(createCommandBuilder()
						.status(CommandStatus.FINISHED)
						.order(Order.VORGANG_LOESCHEN.name())
						.createdAt(getNow().minusMinutes(2)).build());

				var commands = repository.findCommands(Order.VORGANG_LOESCHEN, CommandStatus.FINISHED, getNow().minusSeconds(1));

				assertThat(commands).hasSize(1).first().usingRecursiveComparison().isEqualTo(command);

			}
		}

	}

	@DisplayName("Test adding and updating previous state data")
	@Nested
	class TestHandlingPreviousState {
		private static final String PREVIOUS_STATE_STATUS_VALUE = "NEU";
		private static final String PREVIOUS_STATE_STATUS_FIELD_KEY = Vorgang.MONGODB_FIELDNAME_STATUS;

		private Command command = CommandTestFactory.create();

		@BeforeEach
		void init() {
			command = mongoOperations.save(command);
		}

		@Test
		void shouldSavePreviousStateValue() {
			repository.patch(command.getId(), Map.of(PREVIOUS_STATE_STATUS_FIELD_KEY, PREVIOUS_STATE_STATUS_VALUE));

			var persitedCommand = repository.getById(command.getId());

			assertThat(persitedCommand).isPresent().get()
					.extracting(command -> ((PersistedCommand) command).getPreviousState(), as(InstanceOfAssertFactories.MAP))
					.containsEntry(PREVIOUS_STATE_STATUS_FIELD_KEY,
							PREVIOUS_STATE_STATUS_VALUE);
		}
	}

	@Nested
	class TestDeleteAllByVorgang {
		@Test
		void shouldDeleteCommand() {
			mongoOperations.save(CommandTestFactory.create());

			repository.deleteAllByVorgang(VorgangTestFactory.ID);

			assertThat(mongoOperations.count(query(new Criteria()), Command.COLLECTION_NAME)).isZero();
		}

		@Test
		void shouldNotDeleteOther() {
			mongoOperations.save(CommandTestFactory.create());

			repository.deleteAllByVorgang("other");

			assertThat(mongoOperations.count(query(new Criteria()), Command.COLLECTION_NAME)).isOne();
		}
	}

	@Nested
	class TestExistsCommandWithUserId {

		@BeforeEach
		void setup() {
			mongoOperations.save(CommandTestFactory.create());
		}

		@Test
		void shouldFindInCreatedBy() {
			var isExists = repository.existsCommandWithUserId(CommandTestFactory.CREATED_BY);

			assertThat(isExists).isTrue();
		}

		@Test
		void shouldFindInAssignedTo() {
			var userId = "userId";
			var command = CommandTestFactory.createBuilder().createdBy(null).body(Map.of("assignedTo", userId)).build();
			mongoOperations.save(command);

			var isExists = repository.existsCommandWithUserId(userId);

			assertThat(isExists).isTrue();
		}

		@Test
		void shouldReturnFalse() {
			var esExists = repository.existsCommandWithUserId("userId");

			assertThat(esExists).isFalse();
		}
	}
}