/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.test.context.support.WithMockUser;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
@WithMockUser
class CommandServiceITCase {

	@SpyBean
	private CommandService commandService;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private CurrentUserService currentUserService;

	@Nested
	class TestVorgangCommandPersist {

		@BeforeEach
		void initCurrentUserService() {
			when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));
			when(currentUserService.getUser()).thenReturn(CallContextUserTestFactory.create());
		}

		@BeforeEach
		void persistVorgang() {
			mongoOperations.save(VorgangTestFactory.create());
		}

		@Test
		void persistCommand() {
			Command persistedCommand = commandService.saveCommand(
					CommandTestFactory.createBuilder()
							.relationId(CommandTestFactory.RELATION_ID)
							.relationVersion(CommandTestFactory.RELATION_VERSION).build());

			Optional<Command> commandOpt = commandService.findCommand(persistedCommand.getId());
			assertThat(commandOpt).isPresent();

			Command command = commandOpt.get();
			assertThat(command.getId()).hasSize(36);
			assertThat(command.getCreatedAt()).isNotNull();
			assertThat(command.getStatus()).isEqualTo(CommandStatus.PENDING);
			assertThat(command.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}

		@Test
		void persistCommandHappyPath() {
			CreateCommandRequest request = CreateCommandRequestTestFactory
					.createBuilder()
					.relationId(CommandTestFactory.RELATION_ID)
					.relationVersion(CommandTestFactory.RELATION_VERSION)
					.build();

			Command persistedCommand = commandService.createCommand(request);
			verify(commandService, timeout(500)).setCommandFinished(eq(persistedCommand.getId()), any());

			Optional<Command> commandOpt = commandService.findCommand(persistedCommand.getId());
			assertThat(commandOpt).isPresent();

			Command command = commandOpt.get();
			assertThat(command.getId()).hasSize(36);
			assertThat(command.getCreatedAt()).isNotNull();
			assertThat(command.getCreatedBy()).isEqualTo(request.getCallContext().getUser().getId());
			assertThat(command.getStatus()).isEqualTo(CommandStatus.FINISHED);
			assertThat(command.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}
	}

	@Nested
	class TestVorgangCommandFinished {

		@BeforeEach
		void persistVorgang() {
			mongoOperations.dropCollection(Vorgang.class);
			mongoOperations.dropCollection(Command.class);

			mongoOperations.save(VorgangTestFactory.create());
			mongoOperations.save(CommandTestFactory.createBuilder().relationId(VorgangTestFactory.ID).build());
		}

		@Test
		void setCommandFinishedHappyPath() {
			commandService.setCommandFinished(CommandTestFactory.ID, CommandTestFactory.RELATION_VERSION);

			Command command = mongoOperations.findById(CommandTestFactory.ID, Command.class);

			assertThat(command.getId()).hasSize(36);
			assertThat(command.getCreatedAt()).isNotNull();
			assertThat(command.getCreatedBy()).isNotNull();
			assertThat(command.getStatus()).isEqualTo(CommandStatus.FINISHED);
			assertThat(command.getRelationId()).isEqualTo(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetVorgangCommand {

		private Command command1;
		private Command command2;

		@BeforeEach
		void persistVorgangWithCommands() {
			command1 = mongoOperations.save(CommandTestFactory.createBuilder()
					.id(CommandTestFactory.ID)
					.status(CommandStatus.PENDING)
					.relationId(VorgangTestFactory.ID)
					.build());
			command2 = mongoOperations.save(CommandTestFactory.createBuilder()
					.id(UUID.randomUUID().toString())
					.status(CommandStatus.FINISHED)
					.relationId(VorgangTestFactory.ID)
					.build());
		}

		@Test
		void shouldReturnCommand1() {
			Optional<Command> command = commandService.findCommand(command1.getId());

			assertThat(command).isPresent();
			assertThat(command.get().getId()).isEqualTo(command1.getId());
			assertThat(command.get().getCreatedAt()).isNotNull();
			assertThat(command.get().getCreatedBy()).isNotNull();
			assertThat(command.get().getStatus()).isEqualTo(CommandStatus.PENDING);
			assertThat(command.get().getRelationId()).isEqualTo(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnCommand2() {
			Optional<Command> command = commandService.findCommand(command2.getId());

			assertThat(command).isPresent();
			assertThat(command.get().getId()).isEqualTo(command2.getId());
			assertThat(command.get().getCreatedAt()).isNotNull();
			assertThat(command.get().getCreatedBy()).isNotNull();
			assertThat(command.get().getStatus()).isEqualTo(CommandStatus.FINISHED);
			assertThat(command.get().getRelationId()).isEqualTo(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestExistsPendingCommands {

		@BeforeEach
		void persistVorgang() {
			mongoOperations.dropCollection(Command.class);
		}

		@Test
		void noCommands() {
			assertThat(commandService.existsPendingCommands(VorgangTestFactory.ID)).isFalse();
		}

		@Test
		void onlyFinishedCommands() {
			mongoOperations.save(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).build());
			mongoOperations.save(CommandTestFactory.createBuilder().id("ID2").status(CommandStatus.FINISHED).build());

			assertThat(commandService.existsPendingCommands(VorgangTestFactory.ID)).isFalse();
		}

		@Test
		void pendingCommand() {
			mongoOperations.save(CommandTestFactory.createBuilder().status(CommandStatus.PENDING).build());
			mongoOperations.save(CommandTestFactory.create());

			assertThat(commandService.existsPendingCommands(VorgangTestFactory.ID)).isTrue();
		}

		@Test
		void pendingCommandForAnotherVorgang() {
			mongoOperations.save(CommandTestFactory.createBuilder().vorgangId("ANOTHER_VORGANG_ID").status(CommandStatus.PENDING).build());

			assertThat(commandService.existsPendingCommands(VorgangTestFactory.ID)).isFalse();
		}
	}
}