/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class CommandServiceTest {

	@Spy
	@InjectMocks
	private CommandService service;

	@Mock
	private CommandRepository repository;
	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private CurrentUserService currentUserService;

	@Nested
	class TestCreateCommand {

		private final CreateCommandRequest request = CreateCommandRequestTestFactory.create();

		@Captor
		private ArgumentCaptor<Command> commandCaptor;

		@Test
		void shouldCallOperationRepository() {
			service.createCommand(request);

			verify(repository).save(any());
		}

		@Test
		void shouldReturnCommand() {
			var command = (PersistedCommand) service.createCommand(request);

			assertThat(command.getOrder()).isEqualTo(Order.VORGANG_ANNEHMEN.name());
			assertThat(command.getId()).isNotNull();
			assertThat(command.getStatus()).isEqualTo(CommandStatus.PENDING);
			assertThat(command.getCreatedAt()).isNotNull();
			assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			assertThat(command.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
			assertThat(command.getRelationVersion()).isEqualTo(CreateCommandRequestTestFactory.RELATION_VERSION);
		}

		@Test
		void shouldHandleMissingCallContext() {
			var request = CreateCommandRequestTestFactory.createBuilder().callContext(null).build();

			var command = service.createCommand(request);

			assertThat(command).isNotNull();
		}

		@Nested
		class ShouldAddUser {

			@Test
			void fromCallContext() {
				when(currentUserService.findUser()).thenReturn(Optional.empty());

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			}

			@Test
			void fromUserService() {
				var request = CreateCommandRequestTestFactory.createBuilder().callContext(null).build();
				when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			}

			@Test
			void ifBothArePresentFromUserService() {
				when(currentUserService.findUser())
						.thenReturn(Optional.of(CallContextUserTestFactory.createBuilder().userId(Optional.of("other")).build()));

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo("other");
			}

		}

		@Test
		void shouldHandleEmptyUser() {
			var request = CreateCommandRequestTestFactory.createBuilder().callContext(CallContextTestFactory.createBuilder().user(null).build())
					.build();

			service.createCommand(request);

			verify(repository).save(commandCaptor.capture());
			assertThat(commandCaptor.getValue().getCreatedBy()).isNull();
		}
	}

	@Nested
	class TestGetVorgangCommand {

		final String commandId = CommandTestFactory.ID;

		@BeforeEach
		void mockOperationRepository() {
			when(repository.getById(any())).thenReturn(Optional.of(CommandTestFactory.create()));
		}

		@Test
		void shouldCallOperationRepository() {
			service.findCommand(commandId);

			verify(repository).getById(any());
		}

		@Test
		void shouldReturnCommand() {
			Optional<Command> command = service.findCommand(commandId);

			assertThat(command).isPresent();
			assertThat(command.get().getId()).isEqualTo(commandId);
			assertThat(command.get().getCreatedAt()).isEqualTo(CommandTestFactory.CREATED_AT_STR);
			assertThat(command.get().getCreatedBy()).isEqualTo(CommandTestFactory.CREATED_BY);
			assertThat(command.get().getStatus()).isEqualTo(CommandTestFactory.STATUS);
			assertThat(command.get().getOrder()).isEqualTo(CommandTestFactory.ORDER.name());
			assertThat(command.get().getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}
	}

	@Nested
	class TestFindCommands {

		@Test
		void shouldCallRepository() {
			var createdAfter = ZonedDateTime.now();

			service.findFinishedVorgangLoeschenCommandsOlderThen(createdAfter);

			verify(repository).findCommands(Order.VORGANG_LOESCHEN, CommandStatus.FINISHED, createdAfter);
		}
	}

	@Nested
	class TestSetCommandFinished {

		@Test
		void shouldCallRepositoryUpdateCommandStatus() {
			service.setCommandFinished(CommandTestFactory.ID, 17L);

			verify(repository).updateCommandStatusAndVersion(CommandTestFactory.ID, CommandStatus.FINISHED, 17L);
		}

		@Test
		void shouldCallRepositoryFinishCommand() {
			service.setCommandFinished(CommandTestFactory.ID);

			verify(repository).finishCommand(CommandTestFactory.ID);
		}

		@Test
		void shouldCallRepositoryFinishCommandWithCreatedResource() {
			service.setCommandFinished(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);

			verify(repository).finishCommand(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);
		}
	}

	@Nested
	class TestSetCommandRevoked {

		@Test
		void shouldCallRepositoryUpdateCommandStatus() {
			service.setCommandRevoked(CommandTestFactory.ID);

			verify(repository).updateCommandStatus(CommandTestFactory.ID, CommandStatus.REVOKED);
		}
	}

	@Nested
	class TestRevokeCommandPending {

		final String commandId = CommandTestFactory.ID;

		@Test
		void shouldThrowException() {
			assertThatExceptionOfType(TechnicalException.class).isThrownBy(() -> service.setCommandRevokePending(commandId));
		}

		@Test
		void shouldUpdateCommandStatus() {
			when(repository.getById(commandId)).thenReturn(Optional.of(CommandTestFactory.create()));

			service.setCommandRevokePending(commandId);

			verify(repository).updateCommandStatus(anyString(), any(CommandStatus.class));
		}

		@Test
		void shouldCallPublishRevokeCommandEvent() {
			when(repository.getById(commandId)).thenReturn(Optional.of(CommandTestFactory.create()));

			service.setCommandRevokePending(commandId);

			verify(service).publishRevokeCommandEvent(any(String.class));
		}
	}

	@Nested
	class TestFiringRevokeCommand {

		final String commandId = CommandTestFactory.ID;

		@Test
		void shoudlFireEvent() {
			when(repository.getById(commandId)).thenReturn(Optional.of(CommandTestFactory.create()));

			service.publishRevokeCommandEvent(commandId);

			verify(publisher).publishEvent(any(RevokeCommandEvent.class));
		}

		@Test
		void shoudlThrowTechnicalExceptionBecauseCommandNotFound() {
			when(repository.getById(commandId)).thenReturn(Optional.empty());

			assertThatExceptionOfType(TechnicalException.class).isThrownBy(() -> service.publishRevokeCommandEvent(commandId));
		}

		@Test
		void shoudlThrowTechnicalExceptionBecauseNoPreviousState() {
			when(repository.getById(commandId)).thenReturn(Optional.of(CommandTestFactory.createBuilder().previousState(null).build()));

			assertThatExceptionOfType(TechnicalException.class).isThrownBy(() -> service.publishRevokeCommandEvent(commandId));
		}

	}

	@Nested
	class TestExistsPendingCommands {

		@BeforeEach
		void init() {
			when(repository.existsPendingCommands(any())).thenReturn(true);
		}

		@Test
		void shouldCallRepository() {
			assertThat(service.existsPendingCommands(VorgangTestFactory.ID)).isTrue();

			verify(repository).existsPendingCommands(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		private final List<Command> pendingCommands = Stream.of(CommandTestFactory.create()).collect(Collectors.toList());

		@BeforeEach
		void init() {
			when(repository.getPendingCommands(any())).thenReturn(pendingCommands);
		}

		@Test
		void shouldCallRepository() {
			service.getPendingCommands(VorgangTestFactory.ID);

			verify(repository).getPendingCommands(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var result = service.getPendingCommands(VorgangTestFactory.ID);

			assertThat(result).isEqualTo(pendingCommands);
		}
	}

	@Nested
	class TestCommandSetPreviousState {

		final String commandId = CommandTestFactory.ID;

		@Test
		void shouldSetPreviousState() {
			service.setPreviousState(commandId, CommandTestFactory.PREVIOUS_STATE);

			verify(repository).patch(anyString(), any());
		}
	}

	@Nested
	class TestDeleteAllByVorgang {
		@Test
		void shouldCallRepository() {
			service.deleteAllByVorgang(VorgangTestFactory.ID);

			verify(repository).deleteAllByVorgang(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestFilterUserIdsWithoutCommands {

		@Test
		void shouldCallRepository() {
			service.existsCommandWithUserId(UserTestFactory.ID);

			verify(repository).existsCommandWithUserId(UserTestFactory.ID);
		}

	}
}