/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcEmpty;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcFindCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcOrder;
import de.ozgcloud.vorgang.grpc.command.GrpcRevokeCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcSetCommandExecutedRequest;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class GrpcCommandServiceTest {

	@Spy
	@InjectMocks
	private GrpcCommandService service;
	@Mock
	private GrpcCreateCommandRequestMapper createCommandRequestMapper;
	@Mock
	private CommandService commandService;

	@Mock
	private GrpcCommandResponseMapper commandResponseMapper;
	@Mock
	private GrpcCommandMapper commandMapper;

	@Mock
	private PolicyService policyService;
	@Mock
	private ApplicationEventPublisher eventPublisher;

	@Nested
	class TestCreateCommand {

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommandResponse> responseCaptor;
		private final GrpcCreateCommandRequest request = GrpcCreateCommandRequest.newBuilder().build();
		private final CreateCommandRequest createComandRequest = CreateCommandRequestTestFactory.create();
		private final GrpcCommandResponse response = GrpcCommandResponse.newBuilder().build();
		private final Command persistedCommand = CommandTestFactory.create();

		@BeforeEach
		void mockMapper() {
			when(createCommandRequestMapper.fromGrpc(any(GrpcCreateCommandRequest.class))).thenReturn(createComandRequest);
			when(commandService.createCommand(any(CreateCommandRequest.class))).thenReturn(persistedCommand);
			when(commandService.findCommand(anyString())).thenReturn(Optional.of(CommandTestFactory.create()));
			when(commandResponseMapper.toGrpc(any(CommandResponse.class))).thenReturn(response);
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callCreateCommand();

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallCommandRequestMapper() throws Exception {
			callCreateCommand();

			verify(createCommandRequestMapper).fromGrpc(request);
		}

		@Test
		void shouldCallCommandService() throws Exception {
			callCreateCommand();

			verify(commandService).createCommand(createComandRequest);
		}

		@Test
		void shouldReturnValue() throws Exception {
			callCreateCommand();

			var values = responseCaptor.getValue();

			assertThat(values).isEqualTo(response);
		}

		private void callCreateCommand() throws Exception {
			service.createCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestIsStatusChangeOrder {

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "VORGANG_ANNEHMEN", "VORGANG_VERWERFEN", "VORGANG_ZURUECKHOLEN", "VORGANG_BEARBEITEN",
				"VORGANG_BESCHEIDEN", "VORGANG_ZURUECKSTELLEN", "VORGANG_ABSCHLIESSEN", "VORGANG_WIEDEREROEFFNEN" })
		void shouldReturnTrue(Order order) {
			var result = service.isStatusChangeOrder(order.name());

			assertThat(result).isTrue();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "VORGANG_ANNEHMEN", "VORGANG_VERWERFEN", "VORGANG_ZURUECKHOLEN", "VORGANG_BEARBEITEN",
				"VORGANG_BESCHEIDEN", "VORGANG_ZURUECKSTELLEN", "VORGANG_ABSCHLIESSEN", "VORGANG_WIEDEREROEFFNEN", "VORGANG_ZUM_LOESCHEN_MARKIEREN",
				"VORGANG_LOESCHEN" })
		void shouldReturnFalse(Order order) {
			var result = service.isStatusChangeOrder(order.name());

			assertThat(result).isFalse();
		}
	}

	@DisplayName("Get command")
	@Nested
	class TestGetCommand {

		@Mock
		private StreamObserver<GrpcCommand> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommand> responseCaptor;
		private final GrpcGetCommandRequest request = GrpcGetCommandRequest.newBuilder().setId(CommandTestFactory.ID).build();

		@BeforeEach
		void init() {
			when(commandService.findCommand(any())).thenReturn(Optional.of(CommandTestFactory.create()));
			when(commandMapper.toGrpc(any())).thenReturn(GrpcCommand.newBuilder().setId(CommandTestFactory.ID).build());
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callGetVorgangCommand();

			verify(policyService).checkPermissionByCommand(anyString());
		}

		@Test
		void shouldCallVorgangCommandService() throws Exception {
			callGetVorgangCommand();

			verify(commandService).findCommand(CommandTestFactory.ID);
		}

		@Test
		void shouldCallMapper() throws Exception {
			callGetVorgangCommand();

			verify(commandMapper).toGrpc(any());
		}

		@Test
		void shouldHaveReturnValue() throws Exception {
			callGetVorgangCommand();

			var values = responseCaptor.getValue();

			assertThat(values.getId()).isEqualTo(CommandTestFactory.ID);
		}

		private void callGetVorgangCommand() throws Exception {
			service.getCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestRevokeCommand {

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;
		private final GrpcRevokeCommandRequest request = GrpcRevokeCommandRequest.newBuilder().setId(CommandTestFactory.ID).build();

		@BeforeEach
		void mockCommandService() {
			Command vorgangCommand = CommandTestFactory.createBuilder().id(CommandTestFactory.ID).build();

			when(commandService.findCommand(any())).thenReturn(Optional.of(vorgangCommand));
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callRevokeCommand();

			verify(policyService).checkPermissionByCommand(anyString());
		}

		@Test
		void shouldCallServiceRevokeCommand() throws Exception {
			callRevokeCommand();

			verify(commandService).setCommandRevokePending(request.getId());
		}

		@Test
		void shouldCallMapper() throws Exception {
			callRevokeCommand();

			verify(commandResponseMapper).toGrpc(any());
		}

		@Nested
		class TestProceedRevokeCommand {

			private PersistedCommand command = CommandTestFactory.create();

			@BeforeEach
			void init() {
				when(commandService.findCommand(anyString())).thenReturn(Optional.of(command));
			}

			@Test
			@SneakyThrows
			void shouldLoadCommand() {
				callRevokeCommand();

				verify(service).getCommand(CommandTestFactory.ID);
			}
		}

		private void callRevokeCommand() throws Exception {
			service.revokeCommand(request, responseObserver);
		}
	}

	@Nested
	class TestExistsPendingCommand {

		@Mock
		private StreamObserver<GrpcExistsPendingCommandsResponse> responseObserver;
		private final GrpcExistsPendingCommandsRequest request = GrpcExistsPendingCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID)
				.build();

		@BeforeEach
		void mockCommandService() {
			when(commandService.existsPendingCommands(any())).thenReturn(true);
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			service.existsPendingCommands(request, responseObserver);

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallCommandService() {
			service.existsPendingCommands(request, responseObserver);

			verify(commandService).existsPendingCommands(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		@Mock
		private StreamObserver<GrpcGetPendingCommandsResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetPendingCommandsResponse> responseCaptor;
		private final GrpcGetPendingCommandsRequest request = GrpcGetPendingCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID).build();
		private final Command command = CommandTestFactory.create();

		@BeforeEach
		void mockCommandService() {
			when(commandService.getPendingCommands(anyString())).thenReturn(Stream.of(command).collect(Collectors.toList()));
			when(commandMapper.toGrpc(any())).thenReturn(GrpcCommand.newBuilder().setId(CommandTestFactory.ID).build());
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callGetPendingCommands();

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallCommandService() throws Exception {
			callGetPendingCommands();

			verify(commandService).getPendingCommands(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallMapper() throws Exception {
			callGetPendingCommands();

			verify(commandMapper).toGrpc(command);
		}

		@Test
		void shouldReturnValue() throws Exception {
			callGetPendingCommands();

			var response = responseCaptor.getValue();

			assertThat(response.getCommand(0).getId()).isEqualTo(CommandTestFactory.ID);
		}

		private void callGetPendingCommands() throws Exception {
			service.getPendingCommands(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestFindCommands {

		@Mock
		private StreamObserver<GrpcCommandsResponse> responseObserver;
		private final GrpcFindCommandsRequest.Builder requestBuilder = GrpcFindCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID);

		@Test
		void shouldCallPolicyService() throws Exception {
			callFindCommands(requestBuilder.build());

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallService() {
			callFindCommands(requestBuilder.build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.emptyList(), Optional.empty());
		}

		@Test
		void shouldCallServiceWithStatus() {
			callFindCommands(requestBuilder.addStatus(CommandStatus.FINISHED.name()).build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.singletonList(CommandStatus.FINISHED), Optional.empty());
		}

		@Test
		void shouldCallServiceWithOrder() {
			callFindCommands(requestBuilder.setOrder(GrpcOrder.VORGANG_ABSCHLIESSEN).build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.emptyList(), Optional.of(Order.VORGANG_ABSCHLIESSEN));
		}

		private void callFindCommands(GrpcFindCommandsRequest request) {
			service.findCommands(request, responseObserver);
		}
	}

	@Nested
	class TestSetCommandExecuted {

		@Mock
		private StreamObserver<GrpcEmpty> responseObserver;

		@BeforeEach
		void init() {
			when(commandService.findCommand(any())).thenReturn(Optional.of(CommandTestFactory.create()));
		}

		@Test
		void shouldLoadCommand() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(commandService).findCommand(CommandTestFactory.ID);
		}

		@Test
		void shouldPublishEvent() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(eventPublisher).publishEvent(any(CommandExecutedEvent.class));
		}

		@Test
		void shouldHandleMissingCommand() {
			when(commandService.findCommand(any())).thenReturn(Optional.empty());

			service.setCommandExecuted(buildRequest(), responseObserver);

			verifyNoInteractions(eventPublisher);
		}

		@Test
		void shouldCallOnNext() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(responseObserver).onNext(GrpcEmpty.newBuilder().build());
		}

		@Test
		void shouldCallOnCompleted() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(responseObserver).onCompleted();
		}

		private GrpcSetCommandExecutedRequest buildRequest() {
			return GrpcSetCommandExecutedRequest.newBuilder()
					.setCommandId(CommandTestFactory.ID)
					.build();
		}
	}
}