/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.db;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ConcurrentModificationException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import com.mongodb.Function;
import com.mongodb.client.result.DeleteResult;
import com.mongodb.client.result.UpdateResult;

import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class CollisionVerifierTest {

	private CollisionVerifier verifier;

	@Mock
	private Function<String, Boolean> doEntryWithIdExist;

	@BeforeEach
	void setup() {
		verifier = spy(new CollisionVerifier(doEntryWithIdExist));
	}

	@Nested
	class VerifyUpdateResult {

		@Mock
		private UpdateResult result;

		@Test
		@DisplayName("should detect concurrent modification")
		void shouldThrowConcurrentModification() {
			when(result.getMatchedCount()).thenReturn(0L);
			when(result.getModifiedCount()).thenReturn(0L);
			when(doEntryWithIdExist.apply(anyString())).thenReturn(true);

			assertThatThrownBy(() -> verifier.verify(result, VorgangTestFactory.ID)).isInstanceOf(ConcurrentModificationException.class);
		}

		@Nested
		@DisplayName("should NOT throw concurrent modification exception")
		class NotThrowingException {
			@Test
			void onSuccessfullUpdate() {
				when(result.getModifiedCount()).thenReturn(1L);

				assertThatCode(() -> verifier.verify(result, VorgangTestFactory.ID)).doesNotThrowAnyException();
			}

			@Test
			@DisplayName("if Vorgang does not exists.")
			void onMissingVorgang() {
				when(result.getMatchedCount()).thenReturn(0L);
				when(result.getModifiedCount()).thenReturn(0L);
				when(doEntryWithIdExist.apply(anyString())).thenReturn(false);

				assertThatCode(() -> verifier.verify(result, VorgangTestFactory.ID)).doesNotThrowAnyException();
			}

			@Test
			@DisplayName("on unchanged")
			void onOnchangedVorgang() {
				when(result.getMatchedCount()).thenReturn(1L);
				when(result.getModifiedCount()).thenReturn(0L);

				assertThatCode(() -> verifier.verify(result, VorgangTestFactory.ID)).doesNotThrowAnyException();
			}
		}

	}

	@Nested
	class VerifyDeleteResult {

		@Mock
		private DeleteResult result;

		@BeforeEach
		void setup() {
			when(result.wasAcknowledged()).thenReturn(true);
		}

		@Test
		@DisplayName("should detect concurrent modification")
		void shouldThrowConcurrentModification() {
			when(doEntryWithIdExist.apply(anyString())).thenReturn(true);
			when(result.getDeletedCount()).thenReturn(0L);

			assertThatThrownBy(() -> verifier.verify(result, VorgangTestFactory.ID)).isInstanceOf(ConcurrentModificationException.class);
		}

		@Nested
		@DisplayName("should NOT throw concurrent modification exception")
		class TestNotThrowingException {

			@Test
			void onSuccessfulDelete() {
				when(result.getDeletedCount()).thenReturn(1L);

				assertThatCode(() -> verifier.verify(result, VorgangTestFactory.ID)).doesNotThrowAnyException();
			}

			@Test
			@DisplayName("if VorgangAttachedItem does not exists.")
			void onMissingVorgangAttachedItem() {
				when(doEntryWithIdExist.apply(anyString())).thenReturn(false);
				when(result.getDeletedCount()).thenReturn(0L);

				assertThatCode(() -> verifier.verify(result, VorgangTestFactory.ID)).doesNotThrowAnyException();
			}

		}

	}
}