/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.status.StatusChangedEvent;
import de.ozgcloud.vorgang.vorgang.SetAktenzeichenCompletedEvent;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangAssignedEvent;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEvent;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEventTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class SearchEventListenerTest {

	@InjectMocks
	private SearchEventListener eventListener;

	@Mock
	private SearchService service;
	@Mock
	private CommandService commandService;
	@Mock
	private VorgangService vorgangService;

	private Vorgang vorgang;

	@BeforeEach
	void init() {
		vorgang = VorgangTestFactory.create();
	}

	@Nested
	class TestOnNewVorgang {
		@Mock
		private VorgangCreatedEvent vorgangCreatedEvent;

		@Test
		void onNewVorgangTest() {
			when(vorgangCreatedEvent.getSource()).thenReturn(VorgangTestFactory.ID);
			when(vorgangService.getById(anyString())).thenReturn(vorgang);

			eventListener.onNewVorgang(vorgangCreatedEvent);

			verify(service).addVorgang(any());
		}
	}

	@Nested
	class TestOnVorgangAssigned {
		@Mock
		private VorgangAssignedEvent vorgangAssignedEvent;

		@BeforeEach
		void init() {
			Command command = CommandTestFactory.create();
			when(vorgangAssignedEvent.getCommand()).thenReturn(command);

			when(vorgangService.getById(anyString())).thenReturn(vorgang);
		}

		@Test
		void shouldUpdateVorgang() {
			eventListener.onVorgangAssigned(vorgangAssignedEvent);

			verify(service).updateVorgang(any());
		}
	}

	@Nested
	class TestOnUpdatedVorgang {
		@Mock
		private StatusChangedEvent vorgangUpdatedEvent;

		@BeforeEach
		void init() {
			when(commandService.findCommand(any())).thenReturn(Optional.of(CommandTestFactory.create()));
			when(vorgangService.getById(anyString())).thenReturn(vorgang);
		}

		@Test
		void shouldUpdateVorgangInSearchService() {
			eventListener.onVorgangStatusChanged(vorgangUpdatedEvent);

			verify(service).updateVorgang(any());
		}
	}

	@Nested
	class TestOnDeletedVorgang {

		private VorgangDeletedEvent event = VorgangDeletedEventTestFactory.create();

		@Test
		void shouldCallService() {
			eventListener.onVorgangDeleted(event);

			verify(service).deleteVorgang(any());
		}

		@Test
		void shouldCatchException() {
			doThrow(RuntimeException.class).when(service).deleteVorgang(any());

			assertThatCode(() -> eventListener.onVorgangDeleted(event)).doesNotThrowAnyException();
		}
	}

	@Nested
	class TestOnSetAktenzeichenCompleted {

		private SetAktenzeichenCompletedEvent event;
		private Command command;

		@BeforeEach
		void setup() {
			command = CommandTestFactory.createBuilder().bodyObject(Map.of(VorgangService.BODY_OBJECT_AKTENZEICHEN, VorgangTestFactory.AKTENZEICHEN))
					.build();
			event = new SetAktenzeichenCompletedEvent(command);
		}

		@Test
		void shouldCallService() {
			eventListener.onSetAktenzeichenCompletedEvent(event);

			verify(service).setAktenzeichen(command.getVorgangId(), VorgangTestFactory.AKTENZEICHEN);
		}

	}
}
