/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.dao.DataAccessResourceFailureException;

import de.ozgcloud.vorgang.common.errorhandling.SearchServiceUnavailableException;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequestTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class SearchServiceTest {

	@InjectMocks
	private SearchService searchService;
	@Mock
	private SearchVorgangRepository repository;
	@Mock
	private SearchVorgangCustomRepository searchRepository;

	@Mock
	private IndexedVorgangMapper mapper;

	@DisplayName("Add vorgang")
	@Nested
	class TestAddVorgang {

		@Test
		void shouldCallRepository() {
			searchService.addVorgang(VorgangTestFactory.create());

			verify(mapper).fromVorgang(any());
			verify(repository).save(any());
		}
	}

	@DisplayName("Update vorgang")
	@Nested
	class TestUpdateVorgang {

		@Test
		void shouldCallRepository() {
			searchService.updateVorgang(VorgangTestFactory.create());

			verify(mapper).fromVorgang(any());
			verify(repository).save(any());
		}
	}

	@DisplayName("Exists by id")
	@Nested
	class TestExistsById {

		@Test
		void shouldCallRepositoryExistsById() {
			searchService.existsById(VorgangTestFactory.ID);

			verify(repository).existsById(VorgangTestFactory.ID);
		}
	}

	@DisplayName("Search")
	@Nested
	class TestSearch {

		@Test
		void shouldCallRepositorySearch() {
			searchService.find(FindVorgangRequestTestFactory.create());

			verify(searchRepository).searchBy(any(FindVorgangRequest.class));
		}

		@DisplayName("on unavailable search service")
		@Nested
		class TestSearchServiceUnavailable {

			@BeforeEach
			void mock() {
				when(searchRepository.searchBy(any())).thenThrow(DataAccessResourceFailureException.class);
			}

			@Test
			void shouldThrowException() {
				var request = FindVorgangRequestTestFactory.create();
				assertThatExceptionOfType(SearchServiceUnavailableException.class).isThrownBy(() -> searchService.find(request));
			}
		}
	}

	@Nested
	class TestDeleteVorgang {
		@Test
		void shouldDeleteVorgang() {
			searchService.deleteVorgang(VorgangTestFactory.ID);

			verify(repository).deleteById(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestSetAktenzeichen {

		@Test
		void shouldCallRepository() {
			searchService.setAktenzeichen(VorgangTestFactory.ID, VorgangTestFactory.AKTENZEICHEN);

			verify(searchRepository).update(VorgangTestFactory.ID, Map.of(IndexedVorgang.FIELD_AKTENZEICHEN, VorgangTestFactory.AKTENZEICHEN));
		}
	}

}