/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.elasticsearch.core.ElasticsearchOperations;
import org.springframework.test.context.junit.jupiter.DisabledIf;

import de.ozgcloud.vorgang.vorgang.FilterCriteria;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest;

@DisabledIf(expression = "${elasticTests.disabled:false}", reason = "Wegen hohem Speicherverbrauch des testcontainers")
@SearchITCase
public class SearchVorgangCustomRepositoryITCase {

	@Autowired
	private SearchVorgangCustomRepository repostitory;

	@Autowired
	private ElasticsearchOperations elasticsearchOperations;

	@Nested
	class TestSpecialCharacters {

		private static final String VALUE_PATTERN = "AbC%sTest";
		private static final String SEARCH_BY_PATTERN = "bc%st";

		@BeforeEach
		void init() {
			var indexOperations = elasticsearchOperations.indexOps(IndexedVorgang.class);
			indexOperations.delete();
			indexOperations.create();
			indexOperations.putMapping(IndexedVorgang.class);
		}

		@DisplayName("should find vorgang when vorgangName")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInVorgangName(String character) {
			elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().vorgangName(VALUE_PATTERN.formatted(character)).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first()
					.extracting("name").isEqualTo(VALUE_PATTERN.formatted(character));
		}

		@DisplayName("should find vorgang when vorgangNummer")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInVorgangNummer(String character) {
			elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().vorgangNummer(VALUE_PATTERN.formatted(character)).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first()
					.extracting("nummer").isEqualTo(VALUE_PATTERN.formatted(character));
		}

		@DisplayName("should find vorgang when antragstellerName")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInAntragstellerName(String character) {
			elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().antragstellerName(VALUE_PATTERN.formatted(character)).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1);
		}

		@DisplayName("should find vorgang when antragstellerVorname")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInAntragstellerVorname(String character) {
			elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().antragstellerVorname(VALUE_PATTERN.formatted(character)).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1);
		}

		@DisplayName("should find vorgang when aktenzeichen")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInAktenzeichen(String character) {
			elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().aktenzeichen(VALUE_PATTERN.formatted(character)).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first()
					.extracting("aktenzeichen").isEqualTo(VALUE_PATTERN.formatted(character));
		}

		private FindVorgangRequest searchRequest(String character) {
			return FindVorgangRequest.builder().filterBy(FilterCriteria.builder().build()).searchBy(SEARCH_BY_PATTERN.formatted(character)).build();
		}
	}
}
