/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class BinaryFileRepositoryITCase {

	@Autowired
	private BinaryFileRepository repository;

	@Autowired
	private MongoTemplate template;

	@BeforeEach
	void clearDB() {
		template.dropCollection("fs.files");
		template.dropCollection("fs.chunks");
	}

	@Nested
	class TestUploadFile {
		private UploadedFilesReference ref;
		private OzgFile file;

		@BeforeEach
		void init() {
			ref = UploadedFilesReference.builder()
					.client("client")
					.name("test.file")
					.vorgangId(UUID.randomUUID().toString())
					.build();
			file = OzgFileTestFactory.create();
		}

		@Test
		void shouldStoreFiles() throws IOException {
			FileId id = storeFile(45);

			assertThat(repository.getFileContent(id)).isNotEmpty();
		}

		@Test
		void shouldLoadFile() throws IOException {
			int savedFileMB = 25;
			int fileSizeInBytes = savedFileMB * 1_000_000;

			FileId id = storeFile(savedFileMB);

			InputStream in = repository.getFileContent(id);
			byte[] buffer = in.readAllBytes();

			assertThat(buffer).hasSize(fileSizeInBytes);
		}

		private FileId storeFile(int sizeInMB) throws IOException {
			InputStream stream = IncomingFileTestFactory.createContentStreamInMB(sizeInMB);

			FileId id = repository.addContentStream(ref, file, Optional.empty(), stream);
			stream.close();
			return id;
		}
	}

	@Nested
	class TestFindAllMetaData {

		private FileId fileId;

		@Nested
		class TestSingleMetaData {

			@BeforeEach
			void initData() {
				fileId = initContentStream(OzgFileTestFactory.create());
			}

			@Test
			void shouldReturnResult() {
				var result = callRepository(List.of(fileId.toString()));

				assertThat(result.count()).isEqualTo(1);
			}

			@Test
			void shouldReturnOzgFileData() {
				var result = callRepository(List.of(fileId.toString()));

				var ozgFile = result.toList().get(0);
				assertThat(ozgFile.getId()).isEqualTo(fileId);
				assertThat(ozgFile.getName()).isEqualTo(OzgFileTestFactory.NAME);
				assertThat(ozgFile.getContentType()).isEqualTo(OzgFileTestFactory.CONTENT_TYPE);
				assertThat(ozgFile.getSize()).isEqualTo(OzgFileTestFactory.SIZE);
			}
		}

		@Nested
		class TestMultipleMetaData {

			private FileId otherFileId;

			@BeforeEach
			void initData() {
				fileId = initContentStream(OzgFileTestFactory.create());
				otherFileId = initContentStream(OzgFileTestFactory.create());
			}

			@Test
			void shouldReturnResult() {
				var result = callRepository(List.of(fileId.toString(), otherFileId.toString()));

				assertThat(result.count()).isEqualTo(2);
			}

			@Test
			void shouldReturnOzgFileData() {
				var result = callRepository(List.of(fileId.toString(), otherFileId.toString()));

				var ozgFile = result.toList().get(1);
				assertThat(ozgFile.getId()).isEqualTo(otherFileId);
				assertThat(ozgFile.getName()).isEqualTo(OzgFileTestFactory.NAME);
				assertThat(ozgFile.getContentType()).isEqualTo(OzgFileTestFactory.CONTENT_TYPE);
				assertThat(ozgFile.getSize()).isEqualTo(OzgFileTestFactory.SIZE);
			}

		}

		private FileId initContentStream(OzgFile ozgFile) {
			return repository.addContentStream(UploadedFilesReferenceTestFactory.create(), ozgFile, Optional.of(UserTestFactory.ID),
					new ByteArrayInputStream(OzgFileTestFactory.CONTENT));
		}

		private Stream<OzgFile> callRepository(Collection<String> fileIds) {
			return repository.getAll(fileIds);
		}
	}

	@Nested
	class TestDeleteByVorgang {
		@Test
		void shouldDeleteFiles() {
			var fileId = persistFile();

			repository.deleteAllByVorgang(VorgangTestFactory.ID);

			var loaded = repository.getFile(fileId);
			assertThat(loaded).isNull();
		}

		@Test
		void shouldNotDeleteOtherFiles() {
			var fileId = persistFile();

			repository.deleteAllByVorgang("other");

			var loaded = repository.getFile(fileId);
			assertThat(loaded).isNotNull();
		}

		private FileId persistFile() {
			return repository.addContentStream(UploadedFilesReferenceTestFactory.create(),
					OzgFileTestFactory.create(),
					Optional.of(UserTestFactory.ID),
					new ByteArrayInputStream(OzgFileTestFactory.CONTENT));
		}
	}
}