/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroupTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class EingangFilesRepositoryITCase {

	@Autowired
	private EingangFilesRepository repository;

	@Autowired
	private MongoOperations mongoOperations;

	private final static FileId ID2 = FileId.createNew();

	@Nested
	class TestFindAttachments {

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);
			mongoOperations.save(createVorgang());
		}

		@Test
		void shouldHaveTwoAttachments() {
			List<OzgFile> attachments = repository.findAttachmentsByEingangId(EingangTestFactory.ID);

			assertThat(attachments).hasSize(2);
		}

		@Test
		void validateFirstAttachment() {
			OzgFile attachment = repository.findAttachmentsByEingangId(EingangTestFactory.ID).get(0);

			assertThat(attachment.getId()).isEqualTo(IncomingFileTestFactory.ID);
			assertThat(attachment.getName()).isEqualTo(IncomingFileTestFactory.NAME);
			assertThat(attachment.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
		}

		@Test
		void validateSecondAttachment() {
			OzgFile attachment = repository.findAttachmentsByEingangId(EingangTestFactory.ID).get(1);

			assertThat(attachment.getId()).isEqualTo(ID2);
		}
	}

	@Nested
	class TestFindRepresentations {

		private final Vorgang vorgangWithRepresentations = createVorgang();

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);
			mongoOperations.save(vorgangWithRepresentations);
		}

		@Test
		void shouldHaveTwoRepresentations() {
			List<OzgFile> representations = repository.findRepresentationsByEingangId(EingangTestFactory.ID);

			assertThat(representations).hasSize(2);
		}

		@Test
		void validateFirstRepresentation() {
			OzgFile representation = repository.findRepresentationsByEingangId(EingangTestFactory.ID).get(0);

			assertThat(representation.getId()).isEqualTo(IncomingFileTestFactory.ID);
			assertThat(representation.getName()).isEqualTo(IncomingFileTestFactory.NAME);
			assertThat(representation.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
		}

		@Test
		void validateSecondRepresentation() {
			OzgFile representation = repository.findRepresentationsByEingangId(EingangTestFactory.ID).get(1);

			assertThat(representation.getId()).isEqualTo(ID2);
			assertThat(representation.getName()).isEqualTo(IncomingFileTestFactory.NAME);
			assertThat(representation.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
		}
	}

	private Vorgang createVorgang() {
		return VorgangTestFactory.createBuilder().clearEingangs()
				.eingang(EingangTestFactory.createBuilder().clearAttachments().clearRepresentations()
						.attachments(List.of(IncomingFileGroupTestFactory.createBuilder().clearFiles().files(createTestFiles()).build()))
						.representations(createTestFiles()).build())
				.build();
	}

	private List<IncomingFile> createTestFiles() {
		return List.of(IncomingFileTestFactory.create(),
				IncomingFileTestFactory.createBuilder().id(ID2).build());
	}
}
