/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static de.ozgcloud.vorgang.statistic.CountByPathTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.time.LocalDate;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMapTestFactory;
import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangStubTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

@DataITCase
class StatisticRepositoryITCase {

	@Autowired
	private StatisticRepository statisticRepository;
	@Autowired
	private MongoOperations mongoOperations;

	@Nested
	@DisplayName("Count by status")
	class TestByStatus {

		@Nested
		@DisplayName("with filters")
		class TestFilters {
			private static final String ANOTHER_ORGA_ID = "another-orga-id";

			private Vorgang vorgangInBearbeitung = VorgangTestFactory.createBuilder().id(null).status(Vorgang.Status.IN_BEARBEITUNG).build();

			@BeforeEach
			void setUp() {
				mongoOperations.dropCollection(Vorgang.class);

				var vorgangNeu = VorgangTestFactory.createBuilder().id(null).build();
				var zustaendigeStelle = ZustaendigeStelleTestFactory.createBuilder().organisationseinheitenId(ANOTHER_ORGA_ID).build();
				var eingang = EingangTestFactory.createBuilder().zustaendigeStelle(zustaendigeStelle).build();
				var vorgangNeuAnotherOrgaId = VorgangTestFactory.createBuilder().id(null).clearEingangs().eingang(eingang).build();
				mongoOperations.insertAll(List.of(vorgangNeu, vorgangNeuAnotherOrgaId, vorgangInBearbeitung));
			}

			@Test
			void shouldCountAllVorgang() {
				var statusCounts = statisticRepository.countVorgangByStatus(StatisticFilter.builder().build());

				assertThat(statusCounts).containsOnly(
						CountResultTestFactory.createBuilder().name(Vorgang.Status.NEU.toString()).value(2).build(),
						CountResultTestFactory.createBuilder().name(Vorgang.Status.IN_BEARBEITUNG.toString()).build());
			}

			@Test
			void shouldCountAllVorgangByOrganisatorischeEinheitenIds() {
				var filter = StatisticFilter.builder()
						.organisationsEinheitIds(List.of(ANOTHER_ORGA_ID, ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID)).build();

				var statusCounts = statisticRepository.countVorgangByStatus(filter);

				assertThat(statusCounts).containsOnly(
						CountResultTestFactory.createBuilder().name(Vorgang.Status.NEU.toString()).value(2).build(),
						CountResultTestFactory.createBuilder().name(Vorgang.Status.IN_BEARBEITUNG.toString()).build());
			}

			@Test
			void shouldFilterByOrganisatorischeEinheitenId() {
				var filter = StatisticFilter.builder().organisationsEinheitIds(List.of(ANOTHER_ORGA_ID)).build();

				var statusCounts = statisticRepository.countVorgangByStatus(filter);

				assertThat(statusCounts).containsOnly(CountResult.builder().name(Vorgang.Status.NEU.toString()).value(1).build());
			}

			@Test
			void shouldNotCountInCreation() {
				mongoOperations.dropCollection(Vorgang.class);
				var vorgangInCreation = VorgangTestFactory.createBuilder().id(null).inCreation(true).build();
				mongoOperations.insert(vorgangInCreation);

				var statusCounts = statisticRepository.countVorgangByStatus(StatisticFilter.builder().build());

				assertThat(statusCounts).isEmpty();
			}
		}

		@DisplayName("per status")
		@Nested
		class TestPerStatus {

			@BeforeEach
			void setUp() {
				mongoOperations.dropCollection(Vorgang.class);
			}

			@Test
			void shouldCountAngenommen() {
				initTestWithStatus(Vorgang.Status.ANGENOMMEN);

				var statusCounts = statisticRepository.countVorgangByStatus(StatisticFilter.builder().build());

				assertThat(statusCounts).hasSize(1).first().isEqualTo(buildCountResultForStatus(Vorgang.Status.ANGENOMMEN));
			}

			@Test
			void shouldCountVerworfen() {
				initTestWithStatus(Vorgang.Status.VERWORFEN);

				var statusCounts = countVorgangByStatus();

				assertThat(statusCounts).hasSize(1).first().isEqualTo(buildCountResultForStatus(Vorgang.Status.VERWORFEN));

			}

			@Test
			void shouldCountAbgeschlossen() {
				initTestWithStatus(Vorgang.Status.ABGESCHLOSSEN);

				var statusCounts = countVorgangByStatus();

				assertThat(statusCounts).hasSize(1).first().isEqualTo(buildCountResultForStatus(Vorgang.Status.ABGESCHLOSSEN));

			}

			@Test
			void shouldCountWeitergeleitet() {
				initTestWithStatus(Vorgang.Status.WEITERGELEITET);

				var statusCounts = countVorgangByStatus();

				assertThat(statusCounts).hasSize(1).first().isEqualTo(buildCountResultForStatus(Vorgang.Status.WEITERGELEITET));

			}

			@Test
			void shouldCountBeschieden() {
				initTestWithStatus(Vorgang.Status.BESCHIEDEN);

				var statusCounts = countVorgangByStatus();

				assertThat(statusCounts).hasSize(1).first().isEqualTo(buildCountResultForStatus(Vorgang.Status.BESCHIEDEN));
			}

			private void initTestWithStatus(Vorgang.Status status) {
				var vorgang = VorgangTestFactory.createBuilder().id(null).status(status).build();
				mongoOperations.save(vorgang);
			}

			private CountResult buildCountResultForStatus(Vorgang.Status status) {
				return CountResultTestFactory.createBuilder().name(status.toString()).build();
			}
		}

		private List<CountResult> countVorgangByStatus() {
			return statisticRepository.countVorgangByStatus(StatisticFilter.builder().build());
		}
	}

	@Nested
	class TestByPath {

		private final String FIELD_NAME = "nextFrist";
		private final String FIELD_PATH = "ClientAttribute.%s.%s".formatted(ClientAttributeTestFactory.CLIENT_NAME,
				FIELD_NAME);

		private final static Vorgang DEFAULT_VORGANG = VorgangTestFactory.createBuilder().id(null).build();

		private StatisticFilter statisticFilter = StatisticFilter.builder().build();

		@BeforeEach
		void setup() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			mongoOperations.save(DEFAULT_VORGANG);

			var nextWiedervorlageAttribute = ClientAttributeTestFactory.createBuilder().attributeName(FIELD_NAME).build();
			var vorgangWithNextFrist = VorgangTestFactory.createBuilder().id(null)
					.clientAttributes(ClientAttributesMapTestFactory.createWithAttribute(nextWiedervorlageAttribute))
					.build();
			mongoOperations.save(vorgangWithNextFrist);
		}

		@Test
		void shouldCountVorgangsByPath() {
			var countRequest = CountByPathTestFactory.createBuilder().fieldPath(FIELD_PATH).build();
			var operator = OperatorBuilder.from(GrpcQueryOperator.UNEQUAL).fieldPath(FIELD_PATH).operand(null).build();

			var countEntries = statisticRepository.countVorgangByPath(countRequest, operator, statisticFilter);

			assertThat(countEntries).hasSize(1).first().isEqualTo(CountResultTestFactory.create());
		}

		@Test
		void shouldReturnZeroWhenNotFound() {
			var countRequest = CountByPathTestFactory.createBuilder().fieldPath(FIELD_NAME).build();
			var operator = OperatorBuilder.from(GrpcQueryOperator.UNEQUAL).fieldPath("ClientAttribute." + FIELD_NAME + 1)
					.operand(null).build();

			var countEntries = statisticRepository.countVorgangByPath(countRequest, operator, statisticFilter);

			assertThat(countEntries).hasSize(1).first().isEqualTo(CountResultTestFactory.createBuilder().value(0).build());
		}

		@Test
		void shouldReturnCountVorgangByPathWithCondition() {
			saveVorgangDateClientAttribute(LocalDate.now());
			var countRequest = CountByPathTestFactory.createBuilder().fieldPath(FIELD_PATH).build();
			// TODO hier müssen wir noch mal ran - im Repository-Code dürfen wir keine
			// Abhängigkeit auf gRPC Code haben
			var operator = OperatorBuilder.from(GrpcQueryOperator.LESS_THEN).fieldPath("Vorgang." + FIELD_PATH)
					.operand(LocalDate.now().minusDays(2).toString()).build();

			var countEntries = statisticRepository.countVorgangByPath(countRequest, operator, statisticFilter);

			assertThat(countEntries).hasSize(1).first().usingRecursiveComparison()
					.isEqualTo(CountResultTestFactory.createBuilder().value(3).name(ALIAS).build());
		}

		@Test
		void shouldNotCountDELETEDVorgang() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			mongoOperations.save(VorgangStubTestFactory.create());

			var statistic = statisticRepository.countVorgangByPath(CountByPathTestFactory.create(),
					OperatorTestFactory.createBuilder().fieldPath(CountByPathTestFactory.FIELD_PATH).operand("other").build(), statisticFilter);

			assertThat(statistic).hasSize(1).first().usingRecursiveComparison()
					.isEqualTo(CountResultTestFactory.createBuilder().value(0).build());
		}

		@Test
		void shouldReturnZeroOnEmptyCollection() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);

			var statistic = statisticRepository.countVorgangByPath(CountByPathTestFactory.create(), OperatorTestFactory.create(), statisticFilter);

			assertThat(statistic).hasSize(1).first().usingRecursiveComparison()
					.isEqualTo(CountResultTestFactory.createBuilder().value(0).build());
		}

		private void saveVorgangDateClientAttribute(LocalDate date) {
			var nextWiedervorlageAttribute = ClientAttributeTestFactory.createBuilder().attributeName(FIELD_NAME)
					.stringValue(Optional.of(date.toString())).build();
			var vorgangWithNextFrist = VorgangTestFactory.createBuilder().id(null)
					.clientAttributes(ClientAttributesMapTestFactory.createWithAttribute(nextWiedervorlageAttribute))
					.build();
			mongoOperations.save(vorgangWithNextFrist);
		}
	}
}