/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.commons.lang3.ArrayUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.mongodb.core.query.Criteria;

import de.ozgcloud.vorgang.callcontext.CallContextUser;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.common.operator.OperatorFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

class StatisticServiceTest {

	@Spy
	@InjectMocks
	private StatisticService statisticService;

	@Mock
	private StatisticRepository statisticRepository;
	@Mock
	private OperatorFactory operatorFactory;
	@Mock
	private CurrentUserService currentUserService;
	@Mock
	private CallContextUser user;

	private final VorgangCountRequest request = VorgangCountRequestTestFactory.create();

	@DisplayName("Count vorgang")
	@Nested
	class TestCountVorgang {

		@Mock
		private OperatorBuilder operatorBuilder;

		@BeforeEach
		void setUp() {
			when(currentUserService.getUser()).thenReturn(user);
		}

		@Test
		void shouldCallBuildStatistic() {
			when(operatorBuilder.fieldPath(any())).thenReturn(operatorBuilder);
			when(operatorBuilder.operand(any())).thenReturn(operatorBuilder);
			when(operatorFactory.newUnequalOperatorBuilder()).thenReturn(operatorBuilder);
			statisticService.countVorgang(request);

			verify(statisticService).countVorgang(request);
		}

		@DisplayName("Get organisationsEinheitenIds")
		@Nested
		class TestGetOrganisationsEinheitenIds {

			@BeforeEach
			void setup() {
				lenient().when(operatorBuilder.fieldPath(any())).thenReturn(operatorBuilder);
				lenient().when(operatorBuilder.operand(any())).thenReturn(operatorBuilder);
				lenient().when(operatorFactory.newUnequalOperatorBuilder()).thenReturn(operatorBuilder);
			}

			@DisplayName("should return organisationsEinheitIds when check is necessary")
			@Test
			void shouldReturnOrganisationsEinheitIds() {
				var expectedOrganisationsEinheitIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
				when(user.isOrganisationEinheitenIdCheckNecessary()).thenReturn(true);
				when(user.getOrganisatorischeEinheitenIds()).thenReturn(expectedOrganisationsEinheitIds);

				var organisationsEinheitIds = statisticService.getOrganisationsEinheitIds();

				assertThat(organisationsEinheitIds).isPresent().get().isEqualTo(expectedOrganisationsEinheitIds);
			}

			@DisplayName("should return empty list when check is not necessary")
			@Test
			void shouldReturnEmptyCollection() {
				var organisationsEinheitIds = statisticService.getOrganisationsEinheitIds().get();

				assertThat(organisationsEinheitIds).isEmpty();
			}

			@DisplayName("should return empty optional when check necessary but no organisationsEinheitIds")
			@Test
			void shouldThrowException() {
				when(user.isOrganisationEinheitenIdCheckNecessary()).thenReturn(true);
				when(user.getOrganisatorischeEinheitenIds()).thenReturn(Collections.emptyList());

				var organisationsEinheitIds = statisticService.getOrganisationsEinheitIds();

				assertThat(organisationsEinheitIds).isEmpty();
			}
		}
	}

	@DisplayName("Build statistic")
	@Nested
	class TestBuildStatistic {

		@Mock
		private OperatorBuilder operatorBuilder;

		@BeforeEach
		void setUp() {
			when(currentUserService.getUser()).thenReturn(user);
			when(user.isOrganisationEinheitenIdCheckNecessary()).thenReturn(true);
			when(user.getOrganisatorischeEinheitenIds()).thenReturn(List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));
		}

		@Test
		void shouldSetCountByStatus() {
			var resultsCountByStatus = Map.of(Vorgang.Status.NEU, 1);
			doReturn(resultsCountByStatus).when(statisticService).countVorgangByStatus();
			when(operatorBuilder.fieldPath(any())).thenReturn(operatorBuilder);
			when(operatorBuilder.operand(any())).thenReturn(operatorBuilder);
			when(operatorFactory.newUnequalOperatorBuilder()).thenReturn(operatorBuilder);

			var statistic = statisticService.countVorgang(request);

			assertThat(statistic.getCountByStatus()).isEqualTo(resultsCountByStatus);
		}

		@Test
		void shouldSetCountByPaths() {
			var resultsCountByPath = Map.of(CountResultTestFactory.NAME, 1);
			doReturn(resultsCountByPath).when(statisticService).countVorgangByPathsOld(any());

			var statistic = statisticService.countVorgang(request);

			assertThat(statistic.getCountResult()).isEqualTo(resultsCountByPath);
		}
	}

	@Nested
	class TestFilterByGroupMethod {

		private VorgangStatisticRequest existsRequest = VorgangStatisticRequestTestFactory.createBuilder()
				.statisticGroupMethod(StatisticGroupMethod.EXISTS).build();
		private VorgangStatisticRequest countRequest = VorgangStatisticRequestTestFactory.create();

		@Test
		void shouldReturnCountRequests() {
			var statisticRequests = List.of(countRequest, existsRequest);

			var countRequests = statisticService.getCountRequests(statisticRequests);

			assertThat(countRequests).hasSize(1).first().isEqualTo(countRequest);
		}

		@Test
		void shouldReturnExistsRequests() {
			var statisticRequests = List.of(existsRequest, countRequest);

			var existsRequests = statisticService.getExistsRequests(statisticRequests);

			assertThat(existsRequests).hasSize(1).first().isEqualTo(existsRequest);
		}
	}

	@DisplayName("By status")
	@Nested
	class TestCountByStatus {

		@BeforeEach
		void setUp() {
			when(currentUserService.getUser()).thenReturn(user);
			when(user.isOrganisationEinheitenIdCheckNecessary()).thenReturn(true);
			when(user.getOrganisatorischeEinheitenIds()).thenReturn(List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));
		}

		@Test
		void shouldSetStatusStatistic() {
			var expectedStatus = Vorgang.Status.NEU;
			var countResults = List.of(CountResultTestFactory.createBuilder().name(expectedStatus.toString()).build());
			when(statisticRepository.countVorgangByStatus(any())).thenReturn(countResults);

			var countByStatus = statisticService.countVorgangByStatus();

			assertThat(countByStatus).containsEntry(expectedStatus, CountResultTestFactory.VALUE);
		}

		@Test
		void shouldReturnDefaultValues() {

			var countByStatus = statisticService.countVorgangByStatus();

			assertContainsAllStatusExcepting(countByStatus);
		}

	}

	@Nested
	class TestFillMissingStatusCounter {

		@Test
		void shouldAddMissingCounters() {
			var filledCountByStatus = statisticService.fillMissingStatusCounters(Map.of(Vorgang.Status.NEU, 1));

			assertThat(filledCountByStatus).containsEntry(Vorgang.Status.NEU, 1);
			assertContainsAllStatusExcepting(filledCountByStatus, Vorgang.Status.NEU);
		}

		@Test
		void shouldFillEmptyMap() {
			var countByStatus = statisticService.fillMissingStatusCounters(Collections.emptyMap());

			assertContainsAllStatusExcepting(countByStatus);
		}

	}

	private void assertContainsAllStatusExcepting(Map<Vorgang.Status, Integer> countByStatus, Vorgang.Status... exceptStatus) {
		Arrays.stream(Vorgang.Status.values()).filter(status -> !ArrayUtils.contains(exceptStatus, status))
				.forEach(status -> assertThat(countByStatus).containsEntry(status, 0));
	}

	@DisplayName("By path")
	@Nested
	class TestCountByPath {

		@Mock
		private OperatorBuilder operatorBuilder;

		@BeforeEach
		void setup() {
			when(operatorBuilder.fieldPath(any())).thenReturn(operatorBuilder);
			when(operatorBuilder.operand(any())).thenReturn(operatorBuilder);
			when(operatorFactory.newUnequalOperatorBuilder()).thenReturn(operatorBuilder);
		}

		private final CountByPath countByPath = CountPathRequestTestFactory.create();

		@Test
		void shouldCallRepository() {
			doReturn(Optional.of(Collections.emptyList())).when(statisticService).getOrganisationsEinheitIds();

			statisticService.countVorgangByPathsOld(List.of(countByPath));

			verify(statisticRepository).countVorgangByPath(eq(countByPath), any(), any());
		}

		@Test
		void shouldReturnCountResults() {
			doReturn(Optional.of(Collections.emptyList())).when(statisticService).getOrganisationsEinheitIds();
			when(statisticRepository.countVorgangByPath(any(), any(), any())).thenReturn(List.of(CountResultTestFactory.create()));

			var countResults = statisticService.countVorgangByPathsOld(List.of(countByPath));

			assertThat(countResults).hasSize(1).containsEntry(CountResultTestFactory.NAME, CountResultTestFactory.VALUE);
		}
	}

	@Nested
	@DisplayName("Collect Statistic")
	class TestCollectStatistic {

		private final List<String> expectedOrganisationsEinheitIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);

		@Captor
		private ArgumentCaptor<Set<VorgangStatisticRequest>> statisticCountRequestCaptor;

		@BeforeEach
		void setUp() {
			lenient().when(currentUserService.getUser()).thenReturn(user);
			lenient().when(user.isOrganisationEinheitenIdCheckNecessary()).thenReturn(true);
			lenient().when(user.getOrganisatorischeEinheitenIds()).thenReturn(expectedOrganisationsEinheitIds);
		}

		@Test
		void shouldCallCountVorgangByPaths() {
			var statisticCountRequest = VorgangStatisticRequestTestFactory.create();

			statisticService.collectStatistic(List.of(statisticCountRequest));

			verify(statisticService).countVorgangByPaths(statisticCountRequestCaptor.capture());
			assertThat(statisticCountRequestCaptor.getValue()).hasSize(1).contains(statisticCountRequest);
		}

		@Test
		void shouldReturnStatistic() {
			var statisticCountRequest = VorgangStatisticRequestTestFactory.createBuilder().statisticGroupMethod(StatisticGroupMethod.EXISTS).build();

			statisticService.collectStatistic(List.of(statisticCountRequest));

			verify(statisticService).collectFieldExistsStatistic(statisticCountRequestCaptor.capture());
			assertThat(statisticCountRequestCaptor.getValue()).hasSize(1).contains(statisticCountRequest);

		}

		@Test
		void shouldSetCountByPath() {
			doReturn(StatisticTestFactory.BY_PATH_MAP).when(statisticService).countVorgangByPaths(any());

			var statistic = statisticService.collectStatistic(List.of(VorgangStatisticRequestTestFactory.create()));

			assertThat(statistic.getCountResult()).isEqualTo(StatisticTestFactory.BY_PATH_MAP);
		}

		@Test
		void shouldSetFieldExistsStatistic() {
			var resultMap = Map.of(VorgangStatisticRequestTestFactory.RESULT_NAME, true);
			doReturn(resultMap).when(statisticService).collectFieldExistsStatistic(any());
			doReturn(Collections.emptyMap()).when(statisticService).countVorgangByPaths(any());

			var statistic = statisticService.collectStatistic(List.of(VorgangStatisticRequestTestFactory.create()));

			assertThat(statistic.getExistsResult()).isEqualTo(resultMap);
		}

		@Nested
		class TestCollectFieldExistsStatistic {

			@Captor
			private ArgumentCaptor<StatisticFilter> statisticFilterCaptor;

			@Test
			void shouldCallIsFieldExists() {
				var statisticCountRequest = VorgangStatisticRequestTestFactory.createBuilder().statisticGroupMethod(StatisticGroupMethod.EXISTS)
						.build();

				statisticService.collectFieldExistsStatistic(List.of(statisticCountRequest));

				verify(statisticService).isFieldExists(statisticCountRequest);
			}

			@Test
			void shouldAddOrganisationsEinheitIds() {
				var organisationsEinheitIds = List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
				doReturn(Optional.of(organisationsEinheitIds)).when(statisticService).getOrganisationsEinheitIds();

				statisticService.isFieldExists(VorgangStatisticRequestTestFactory.create());

				verify(statisticRepository).isFieldsExists(statisticFilterCaptor.capture());
				assertThat(statisticFilterCaptor.getValue()).extracting("organisationsEinheitIds").isEqualTo(organisationsEinheitIds);
			}

			@Test
			void shouldAddFilterCriteria() {
				var expectedCriteria = Criteria.where("field").exists(true);
				Operator operator = when(mock(Operator.class).getCriteria()).thenReturn(expectedCriteria).getMock();
				var statisticRequest = VorgangStatisticRequestTestFactory.createBuilder().operator(operator).build();

				statisticService.isFieldExists(statisticRequest);

				verify(statisticRepository).isFieldsExists(statisticFilterCaptor.capture());
				assertThat(statisticFilterCaptor.getValue()).extracting("filterCriteria").isEqualTo(expectedCriteria);
			}

		}

		@Nested
		@DisplayName("Count Vorgang by paths")
		class TestCountVorgangByPaths {

			@Captor
			private ArgumentCaptor<StatisticFilter> statisticFilterArgumentCaptor;

			@BeforeEach
			void setUp() {
				lenient().doReturn(Optional.of(List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID))).when(statisticService)
						.getOrganisationsEinheitIds();
			}

			@Test
			void shouldCallCountVorgangByPathForRequest() {
				var vorgangStatisticRequest = VorgangStatisticRequestTestFactory.create();

				statisticService.countVorgangByPaths(List.of(vorgangStatisticRequest));

				verify(statisticService).countVorgangByPath(vorgangStatisticRequest);
			}

			@Test
			void shouldSetOrganisationsEinheitIds() {
				var statisticRequest = VorgangStatisticRequestTestFactory.create();

				statisticService.countVorgangByPath(statisticRequest);

				verify(statisticRepository).countVorgangByPath(any(), any(), statisticFilterArgumentCaptor.capture());
				assertThat(statisticFilterArgumentCaptor.getValue()).extracting("organisationsEinheitIds").isEqualTo(expectedOrganisationsEinheitIds);
			}

			@Test
			void shouldReturnCountResults() {
				doReturn(Stream.of(CountResultTestFactory.create())).when(statisticService).countVorgangByPath(any());

				var countResults = statisticService.countVorgangByPaths(List.of(VorgangStatisticRequestTestFactory.create()));

				assertThat(countResults).hasSize(1).containsEntry(CountResultTestFactory.NAME, CountResultTestFactory.VALUE);
			}
		}
	}
}