/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.status;

import static org.assertj.core.api.Assertions.*;

import java.util.ConcurrentModificationException;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class StatusRepositoryITCase {

	@Autowired
	private StatusRepository repository;
	@Autowired
	private MongoOperations mongoOperations;

	@Nested
	class TestUpdateStatus {

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);

			mongoOperations.save(VorgangTestFactory.create());
		}

		@Test
		void shouldUpdateVorgangStatus() {
			repository.patch(VorgangTestFactory.ID, VorgangTestFactory.VERSION, Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Status.IN_BEARBEITUNG));

			assertThat(mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class).getStatus()).isEqualTo(Status.IN_BEARBEITUNG);
		}

		@Test
		void shouldUpdateVorgangVersion() {
			repository.patch(VorgangTestFactory.ID, VorgangTestFactory.VERSION, Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Status.ANGENOMMEN));

			assertThat(mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class).getVersion()).isEqualTo(VorgangTestFactory.VERSION + 1);
		}

		@Test
		void shouldNotUpdateOnVersionMismatch() {
			Map<String, Object> patch = Map.of(Vorgang.MONGODB_FIELDNAME_STATUS, Status.VERWORFEN.toString());

			assertThatThrownBy(() -> repository.patch(VorgangTestFactory.ID, VorgangTestFactory.VERSION + 5, patch))
					.isInstanceOf(ConcurrentModificationException.class);

			var vorgang = mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class);
			assertThat(vorgang.getVersion()).isEqualTo(VorgangTestFactory.VERSION);
			assertThat(vorgang.getStatus()).isEqualTo(VorgangTestFactory.STATUS);
		}
	}

	@Nested
	class TestUpdateStatusWithoutVersionCheck {

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(Vorgang.class);

			mongoOperations.save(VorgangTestFactory.create());
		}

		@Test
		void shouldUpdateVorgangStatus() {
			repository.updateStatusWithoutVersionCheck(VorgangTestFactory.ID, Status.VERWORFEN);

			assertThat(mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class).getStatus()).isEqualTo(Status.VERWORFEN);
		}

		@Test
		void shouldUpdateVorgangVersion() {
			repository.updateStatusWithoutVersionCheck(VorgangTestFactory.ID, Status.VERWORFEN);

			assertThat(mongoOperations.findById(VorgangTestFactory.ID, Vorgang.class).getVersion()).isEqualTo(VorgangTestFactory.VERSION + 1);
		}
	}

	@Nested
	class TestStatusChangeWithCollision {

		private final Map<String, Object> STATUS_PATCH = Map.of("status", Vorgang.Status.BESCHIEDEN.toString());

		@Test
		void shouldThrowException() {
			mongoOperations.save(VorgangTestFactory.create());

			assertThatThrownBy(() -> repository.patch(VorgangTestFactory.ID, VorgangTestFactory.VERSION - 1, STATUS_PATCH))
					.isInstanceOf(ConcurrentModificationException.class);
		}
	}
}
