/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.user;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.user.grpc.userprofile.UserProfileServiceGrpc.UserProfileServiceBlockingStub;
import de.ozgcloud.user.userprofile.GrpcDeleteInactiveUserRequest;
import de.ozgcloud.user.userprofile.GrpcFindInactiveUserIdsResponse;
import de.ozgcloud.user.userprofile.GrpcUserProfileId;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;

class UserProfileRemoteServiceTest {

	@Spy
	@InjectMocks
	private UserProfileRemoteService service;

	@Mock
	private UserProfileServiceBlockingStub userProfileServiceStub;

	private static final GrpcFindInactiveUserIdsResponse RESPONSE = GrpcFindInactiveUserIdsResponse.newBuilder()
			.addUserProfileIds(GrpcUserProfileId.newBuilder().setId(UserTestFactory.ID)).build();

	@Nested
	class TestFindInactiveUserIds {

		@BeforeEach
		void setup() {
			doReturn(userProfileServiceStub).when(service).getUserServiceStub();
			when(userProfileServiceStub.findInactiveUserIds(any())).thenReturn(RESPONSE);
		}

		@Test
		void shouldCallUserServiceStub() {
			service.findInactiveUserIds();

			verify(userProfileServiceStub).findInactiveUserIds(any());
		}

		@Test
		void shouldMapResponse() {
			service.findInactiveUserIds();

			verify(service).mapFindInactiveUserIdsResponse(RESPONSE);
		}
	}

	@Nested
	class TestResponseMapping {

		@Test
		void shouldReturnUserIds() {
			var result = service.mapFindInactiveUserIdsResponse(RESPONSE);

			assertThat(result).containsExactly(UserTestFactory.ID);
		}
	}

	@Nested
	class TestDeleteInactiveUser {

		@BeforeEach
		void setup() {
			doReturn(userProfileServiceStub).when(service).getUserServiceStub();
		}

		@Test
		void shouldCallUserServiceStub() {
			var deleteRequest = GrpcDeleteInactiveUserRequest.newBuilder().setUserId(UserTestFactory.ID).build();
			doReturn(deleteRequest).when(service).buildDeleteInactiveUsersRequest(any());

			service.deleteInactiveUser(UserTestFactory.ID);

			verify(userProfileServiceStub).deleteInactiveUser(deleteRequest);
		}

		@Test
		void shouldCallBuildRequest() {
			service.deleteInactiveUser(UserTestFactory.ID);

			verify(service).buildDeleteInactiveUsersRequest(UserTestFactory.ID);
		}
	}

	@Nested
	class TestBuildDeleteInactiveUsersRequest {

		@Test
		void shouldBuildRequest() {
			var result = service.buildDeleteInactiveUsersRequest(UserTestFactory.ID);

			assertThat(result.getUserId()).isEqualTo(UserTestFactory.ID);
		}
	}

	@Nested
	class TestGetInterceptor {

		@Test
		void shouldCallSetInterceptor() {
			service.getUserServiceStub();

			verify(userProfileServiceStub).withInterceptors(any());
		}
	}
}
