/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMap;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMapTestFactory;
import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.inProcess.address=in-process:test"
})
@DirtiesContext
@DataITCase
class GrpcVorgangServiceFindQueryITCase {

	private static final String VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_PATH = "ClientAttribute.OzgCloud_NachrichtenManager.hasNewPostfachNachricht";

	@Autowired
	private VorgangHeaderRepository repository;
	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private CurrentUserService currentUserService;

	@GrpcClient("inProcess")
	private VorgangServiceGrpc.VorgangServiceBlockingStub vorgangServiceStub;

	private Vorgang vorgangWithNextFristMinus2Days;
	private Vorgang vorgangNewCreatedAt;

	@BeforeEach
	void init() {
		when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));
		when(currentUserService.getUser()).thenReturn(CallContextUserTestFactory.create());

		mongoOperations.dropCollection(Vorgang.class);
		mongoOperations.save(VorgangTestFactory.createBuilder().id(null)
				.createdAt(ZonedDateTime.now())
				.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now())).build());
		vorgangWithNextFristMinus2Days = mongoOperations.save(VorgangTestFactory.createBuilder().id(null)
				.status(Vorgang.Status.BESCHIEDEN)
				.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minusDays(2))).build());
		vorgangNewCreatedAt = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
	}

	@Test
	void shouldFindWhereExists() {
		var request = GrpcFindVorgangRequestTestFactory.createBuilder().clearFilterBy().setQuery(
				GrpcQuery.newBuilder()
						.addExpressions(GrpcVorgangQueryExpression.newBuilder()
								.setOperator(GrpcQueryOperator.EXISTS)
								.setPath("ClientAttribute.%s.nextWiedervorlageFrist".formatted(CriteriaUtil.DEFAULT_CLIENT)))).build();

		var result = vorgangServiceStub.findVorgang(request);

		assertThat(result.getVorgangList()).hasSize(2);
	}

	@Test
	void shouldFindWithCertainNextFrist() {
		var request = GrpcFindVorgangRequestTestFactory.createBuilder()
				.clearFilterBy()
				.setOrderBy(GrpcFindVorgangRequest.GrpcOrderBy.CREATED_AT_DESC)
				.setQuery(GrpcQuery.newBuilder().addExpressions(
						GrpcVorgangQueryExpression.newBuilder()
								.setOperator(GrpcQueryOperator.LESS_THEN)
								.setPath("ClientAttribute.%s.nextWiedervorlageFrist".formatted(CriteriaUtil.DEFAULT_CLIENT))
								.setOperandStringValue(LocalDate.now().minusDays(1).toString())))
				.build();

		var result = vorgangServiceStub.findVorgang(request);

		assertThat(result.getVorgangList()).hasSize(1).first().extracting(GrpcVorgangHeader::getId).isEqualTo(vorgangWithNextFristMinus2Days.getId());
	}

	@Test
	void shouldFindByComplexCondition() {
		var request = GrpcFindVorgangRequestTestFactory.createBuilder().clearFilterBy()
				.setOrderBy(GrpcFindVorgangRequest.GrpcOrderBy.CREATED_AT_DESC)
				.setQuery(GrpcQuery.newBuilder()
						.addExpressions(GrpcVorgangQueryExpression.newBuilder()
								.setOperator(GrpcQueryOperator.EQUAL)
								.setPath("Vorgang." + Vorgang.MONGODB_FIELDNAME_STATUS)
								.setOperandStringValue("NEU"))
						.addExpressions(GrpcVorgangQueryExpression.newBuilder()
								.setOperator(GrpcQueryOperator.LESS_THEN_OR_EQUAL_TO)
								.setPath("Vorgang." + Vorgang.MONGODB_FIELDNAME_CREATED_AT)
								.setOperandStringValue(VorgangTestFactory.CREATED_AT.toString()))
				)
				.build();

		var result = vorgangServiceStub.findVorgang(request);

		assertThat(result.getVorgangList()).hasSize(1).first().extracting(GrpcVorgangHeader::getId).isEqualTo(vorgangNewCreatedAt.getId());
	}

	@Test
	void shouldFindVorgangWithNewPostfachNachricht() {
		var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).clientAttributes(createClientAttributesMap()).build());

		var request = GrpcFindVorgangRequestTestFactory.createBuilder().clearFilterBy()
				.setOrderBy(GrpcFindVorgangRequest.GrpcOrderBy.CREATED_AT_DESC)
				.setQuery(GrpcQuery.newBuilder().addExpressions(
						GrpcVorgangQueryExpression.newBuilder()
								.setPath(VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_PATH)
								.setOperator(GrpcQueryOperator.UNEQUAL)))
				.build();

		var result = vorgangServiceStub.findVorgang(request);

		assertThat(result.getVorgangList()).hasSize(1).first().extracting(GrpcVorgangHeader::getId).isEqualTo(vorgang.getId());
	}

	private ClientAttributesMap createClientAttributesMapWithWiedervorlageNextFrist(LocalDate nextFrist) {
		return ClientAttributesMapTestFactory.createWithAttribute(buildWiedervorlageNextFristClientAttribute(nextFrist));
	}

	private ClientAttribute buildWiedervorlageNextFristClientAttribute(LocalDate nextFrist) {
		return ClientAttributeTestFactory.createBuilder()
				.clientName(CriteriaUtil.DEFAULT_CLIENT)
				.attributeName(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST)
				.stringValue(Optional.of(nextFrist.toString())).build();
	}

	private ClientAttributesMap createClientAttributesMap() {
		return ClientAttributesMapTestFactory.createWithAttribute(buildClientAttribute());
	}

	private ClientAttribute buildClientAttribute() {
		return ClientAttributeTestFactory.createBuilder()
				.clientName("OzgCloud_NachrichtenManager")
				.attributeName("hasNewPostfachNachricht")
				.intValue(Optional.empty())
				.boolValue(Optional.of(true))
				.build();
	}
}
