/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeMap;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMap;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMapTestFactory;
import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.vorgang.FindVorgangQuery.LogicalOperator;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest.OrderCriteria;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;

@DataITCase
class VorgangHeaderRepositoryITCase {

	@Autowired
	private VorgangHeaderRepository repository;
	@Autowired
	private MongoOperations mongoOperations;

	@BeforeEach
	void clearVorgaenge() {
		mongoOperations.dropCollection(Vorgang.class);
	}

	@Nested
	class TestSortByPriority {

		final FindVorgangRequest request = FindVorgangRequestTestFactory.createBuilder()
				.orderBy(OrderCriteria.PRIORITY)
				.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().build())
				.searchBy(StringUtils.EMPTY).build();

		@Test
		void shouldSortExpiredWiedervorlagenByFrist() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(2, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(3, ChronoUnit.DAYS))).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		// based on OZG-1991
		@DisplayName("should sort vorgaenge with expired wiedervorlage by frist even if another wiedervorlage is done")
		@Test
		void shouldSortExpiredWiedervorlagenByFristWithDoneWidervorlage() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(3, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(2, ChronoUnit.DAYS))).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(3).extracting(VorgangHeader::getId).containsExactly("1", "3", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG" })
		void shouldSortZuBearbeitenAfterExpiredWdvlg(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(status).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(2, ChronoUnit.DAYS))).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@Test
		void shouldSortOpenWdvlgAfterZuBearbeiten() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(3, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(2, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.NEU).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(3).extracting(VorgangHeader::getId).containsExactly("1", "2", "3");
		}

		@Test
		void shouldSortZurBearbeitungByCreatedAt() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.NEU)
					.createdAt(ZonedDateTime.now()).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.IN_BEARBEITUNG)
					.createdAt(ZonedDateTime.now().minus(1, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@Test
		void shouldSortOpendWdvlgIgnoringClosedWdvlg() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS))
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(4, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3").createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());

			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").createdAt(ZonedDateTime.now().minus(1, ChronoUnit.DAYS))
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(3, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("4").createdAt(ZonedDateTime.now().minus(1, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(4).extracting(VorgangHeader::getId).containsExactly("3", "4", "1", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "BESCHIEDEN", "WEITERGELEITET" })
		void shouldSortBearbeitetAfterWdvlg(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(2, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(status).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "BESCHIEDEN", "WEITERGELEITET" })
		void shouldReturnBearbeitetOnlyOnce(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(status).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(1);
		}

		// Zeigt den Bug von OZG-3156
		@ParameterizedTest
		@EnumSource
		void shouldReturnWithWdvlgOnlyOnce(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(status)
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(2, ChronoUnit.DAYS))).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(1);
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "WEITERGELEITET" })
		void shouldContainRemainingStatus(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(status).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(1);
		}

		@Test
		void shouldPageableResultByGivenLimit() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(String.valueOf("1")).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id(String.valueOf("2")).build());

			var request = FindVorgangRequestTestFactory.createBuilder()
					.orderBy(OrderCriteria.PRIORITY)
					.filterBy(null).searchBy(StringUtils.EMPTY)
					.limit(1)
					.offset(0)
					.build();
			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(1);
		}

		@Test
		void shouldConsiderWiedervorlageDone() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(2, ChronoUnit.DAYS))).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2);
			assertThat(vorgaenge.get(0).getId()).isEqualTo("2");
			assertThat(vorgaenge.get(1).getId()).isEqualTo("1");
		}

		// based on OZG-1822
		@Test
		void shouldShowVorgangWithDoneAndOpenWiedervorlage() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(1, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("2", "1");
		}

		// based on OZG-1826, OZG-1885
		@DisplayName("should show Vorgang with expired Wdvlg only once.")
		@ParameterizedTest
		@EnumSource
		void shouldShowExpiredWdvlgOnlyOnce(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.NEU).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(status)
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(1, ChronoUnit.DAYS)))
					.build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2);
		}
	}

	@DisplayName("Sort createdAt desc")
	@Nested
	class TestSortCreatedAtDesc {

		@DisplayName("with single status filter criteria")
		@Nested
		class TestWithSingleStatusCriteria {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(3, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("3").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("4").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(5, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("5").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(6, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("6").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(7, ChronoUnit.DAYS)).build());
			}

			@Test
			void shouldReturnSortedVorgaenge() {
				var request = buildRequestWithFilterCriteria(FilterCriteriaTestFactory.createBuilder().assignedTo(null).build());

				var vorgaenge = repository.findAll(request).getContent();

				assertThat(vorgaenge).hasSize(6).extracting(VorgangHeader::getId).containsExactly("1", "2", "3", "4", "5", "6");
			}
		}

		@DisplayName("with single status and assignedTo filter criteria")
		@Nested
		class TestWithSingleStatusAndAssignedToCriteria {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(3, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("3").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS))
						.assignedTo(null)
						.build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("4").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(5, ChronoUnit.DAYS)).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("5").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(6, ChronoUnit.DAYS))
						.assignedTo(null)
						.build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("6").status(Status.NEU)
						.createdAt(ZonedDateTime.now().minus(7, ChronoUnit.DAYS)).build());
			}

			@Test
			void shouldReturnSortedVorgaenge() {
				var request = buildRequestWithFilterCriteria(FilterCriteriaTestFactory.create());

				var vorgaenge = repository.findAll(request).getContent();

				assertThat(vorgaenge).hasSize(4).extracting(VorgangHeader::getId).containsExactly("1", "2", "4", "6");
			}
		}

		private FindVorgangRequest buildRequestWithFilterCriteria(FilterCriteria filterCriteria) {
			return FindVorgangRequestTestFactory.createBuilder()
					.orderBy(OrderCriteria.CREATED_AT_DESC)
					.filterBy(filterCriteria)
					.searchBy(StringUtils.EMPTY).build();
		}
	}

	@DisplayName("Sort hasNextWiedervorlageFrist")
	@Nested
	class TestSortHasNextWiedervorlageFrist {

		@Test
		void shouldReturnSortedVorgaenge() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now())).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(1, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(1, ChronoUnit.DAYS))).build());

			var request = buildRequestWithFilterCriteria(FilterCriteria.builder().build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(3).extracting(VorgangHeader::getId).containsExactly("3", "2", "1");
		}

		private FindVorgangRequest buildRequestWithFilterCriteria(FilterCriteria filterCriteria) {
			return FindVorgangRequestTestFactory.createBuilder()
					.orderBy(OrderCriteria.NEXT_WIEDERVORLAGE_FRIST)
					.filterBy(filterCriteria)
					.searchBy(StringUtils.EMPTY).build();
		}
	}

	@Nested
	class TestSortForEA {

		final FindVorgangRequest request = FindVorgangRequestTestFactory.createBuilder().offset(0).limit(10)
				.orderBy(OrderCriteria.EA_PRIORITY)
				.filterBy(null).searchBy(StringUtils.EMPTY).build();

		@Test
		void shouldSortInBearbeitungByCreatedAt() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3").status(Status.IN_BEARBEITUNG)
					.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.IN_BEARBEITUNG)
					.createdAt(ZonedDateTime.now().minus(3, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.IN_BEARBEITUNG)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(3).extracting(VorgangHeader::getId).containsExactly("1", "2", "3");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "VERWORFEN" })
		void shouldSortAbzuarbeitendeAfterInBearbeitung(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(status)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.IN_BEARBEITUNG)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG" })
		void shouldSortAbzuarbeitendeByCreatedAt(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(status)
					.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(status)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG" })
		void shouldSortWeitergeleitetAfterAbzuarbeitende(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.WEITERGELEITET)
					.createdAt(ZonedDateTime.now().minus(5, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(status)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@Test
		void shouldSortWeitergeleitetByCreatedAt() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(Status.WEITERGELEITET)
					.createdAt(ZonedDateTime.now().minus(5, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.WEITERGELEITET)
					.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "NEU", "ANGENOMMEN", "VERWORFEN", "WEITERGELEITET", "IN_BEARBEITUNG" })
		void shouldSortRemainingStatusAfterWeitergeleitet(Status status) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").status(status)
					.createdAt(ZonedDateTime.now().minus(4, ChronoUnit.DAYS)).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").status(Status.WEITERGELEITET)
					.createdAt(ZonedDateTime.now().minus(2, ChronoUnit.DAYS)).build());

			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(2).extracting(VorgangHeader::getId).containsExactly("1", "2");
		}

		@Test
		void shouldHaveTotalSize() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());

			var page = repository.findAll(PageRequest.of(0, 10));

			assertThat(page.getTotalElements()).isEqualTo(3);
			assertThat(page.getTotalPages()).isEqualTo(1);
		}

		@Test
		void shouldPageableResultByGivenLimit() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1").build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2").build());

			var request = FindVorgangRequestTestFactory.createBuilder()
					.limit(1)
					.orderBy(OrderCriteria.EA_PRIORITY)
					.filterBy(null)
					.searchBy(StringUtils.EMPTY)
					.build();
			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(1);
			assertThat(vorgaenge.get(0).getId()).isEqualTo("1");
		}

		@Test
		void shouldHaveVorgaengeIgnoringWdvl() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(3, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("2")
					.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().minus(3, ChronoUnit.DAYS))).build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id("3").build());

			var request = FindVorgangRequestTestFactory.createBuilder()
					.orderBy(OrderCriteria.EA_PRIORITY)
					.filterBy(null)
					.searchBy(StringUtils.EMPTY)
					.build();
			var vorgaenge = repository.findAll(request).getContent();

			assertThat(vorgaenge).hasSize(3).extracting(VorgangHeader::getId).containsExactly("1", "2", "3");
		}
	}

	@Nested
	class TestFilterByOrganisationseinheit {

		private final String organisationseinheitenId = UUID.randomUUID().toString();
		private final FindVorgangRequest request = FindVorgangRequestTestFactory.createBuilder().searchBy(StringUtils.EMPTY).build();

		@BeforeEach
		void prepareDatabase() {
			mongoOperations.save(buildWithOrganisationseinheitId(organisationseinheitenId));
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
		}

		@Nested
		class VorgangHeaderPage {

			@Test
			void shouldHaveTotalElements() {
				var vorgangHeaderPage = repository.findAll(request);

				assertThat(vorgangHeaderPage.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldHaveNumberOfElements() {
				var vorgangHeaderPage = repository.findAll(request);

				assertThat(vorgangHeaderPage.getNumberOfElements()).isEqualTo(1);
			}

			@Test
			void shouldHaveTotalPages() {
				var vorgangHeaderPage = repository.findAll(request);

				assertThat(vorgangHeaderPage.getTotalPages()).isEqualTo(1);
			}
		}

		@Nested
		class TestVorgangWithWiedervorlagen {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.dropCollection(Vorgang.class);

				mongoOperations.save(buildWithOrganisationseinheitId(organisationseinheitenId));
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null)
						.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(2, ChronoUnit.DAYS)))
						.build());
			}

			@Test
			void shouldFilterByDefaultId() {
				var vorgangHeaderPage = repository.findAll(request);

				assertThat(vorgangHeaderPage.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldFilterByGivenId() {
				var filterCriteria = FilterCriteriaTestFactory.createBuilder().clearOrganisationseinheitIds()
						.organisationseinheitId(organisationseinheitenId).build();
				var vorgangHeaderPage = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(vorgangHeaderPage.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldHaveJustOneVorgangStatusNeu() {
				mongoOperations.dropCollection(Vorgang.class);
				mongoOperations.save(VorgangTestFactory.createBuilder()
						.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(2, ChronoUnit.DAYS))).build());

				var vorgangHeaderPage = repository.findAll(request);

				assertThat(vorgangHeaderPage.getTotalElements()).isEqualTo(1);
			}
		}
	}

	@Nested
	class TestFilterBy {

		@Nested
		class TestInCreation {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.save(VorgangTestFactory.createBuilder().inCreation(true).build());
			}

			@Test
			void shouldIgnoreVorgangInCreation() {
				var filterCriteria = FilterCriteriaTestFactory.create();

				var page = callRepositoryFindAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isZero();
			}
		}

		@Nested
		class TestOrganisationseinheitenId {

			@BeforeEach
			void prepareDatabase() {
				mongoOperations.save(buildWithOrganisationseinheitId("not matched"));
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			}

			@DisplayName("should not call database and return an empty list if filterby ist true with an empty list")
			@Test
			void shouldReturnEmptyPage() {
				var filterCriteria = FilterCriteria.builder().filterByOrganisationseinheitenId(true).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isZero();
			}

			@DisplayName("should return filtered page if filter for organisationseinheitenid is true and list is not empty")
			@Test
			void shouldReturnFilteredPage() {
				var filterCriteria = FilterCriteriaTestFactory.createBuilder().clearStatus().build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(1);
			}

			@DisplayName("should return not filtered page if filter for organisationseinheitenid is false")
			@Test
			void shouldReturnNotFilteredPage() {
				var filterCriteria = FilterCriteria.builder().filterByOrganisationseinheitenId(false).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(2);
			}

			@DisplayName("should return not filtered page if filter for organisationseinheitenid is false and list is not empty")
			@Test
			void shouldReturnNotFilteredPageOnFilterFalse() {
				var filterCriteria = FilterCriteriaTestFactory.createBuilder().clearStatus().filterByOrganisationseinheitenId(false).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(2);
			}
		}

		@Nested
		class TestStatus {

			private final FilterCriteria.FilterCriteriaBuilder filterCriteriaBuilder = FilterCriteria.builder()
					.filterByOrganisationseinheitenId(false);

			@BeforeEach
			void mockDatabase() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.NEU).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.IN_BEARBEITUNG).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.ANGENOMMEN).build());
			}

			@Test
			void shouldFilterBySingleVorgangStatus() {
				var filterCriteria = filterCriteriaBuilder.singleStatus(Status.NEU).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldFilterByVorgangStatus() {
				var filterCriteria = filterCriteriaBuilder.status(List.of(Status.NEU, Status.IN_BEARBEITUNG)).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(2);
			}
		}

		@Nested
		class TestAssignedTo {

			private final FilterCriteria.FilterCriteriaBuilder filterCriteriaBuilder = FilterCriteriaTestFactory.createBuilder()
					.filterByOrganisationseinheitenId(false)
					.clearStatus()
					.clearOrganisationseinheitIds();

			@BeforeEach
			void initVorgaenge() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).assignedTo(null).build());
			}

			@Test
			void shouldReturnFilteredVorgangList() {
				var filterCriteria = filterCriteriaBuilder.assignedTo(UserTestFactory.ID).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldReturnVorgang() {
				var filterCriteria = filterCriteriaBuilder.assignedTo(UserTestFactory.ID).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getContent().get(0)).usingRecursiveComparison()
						.ignoringFields("id", Vorgang.FIELD_CLIENT_ATTRIBUTES)
						.isEqualTo(VorgangTestFactory.create());
			}

			@DisplayName("should return an unfiltered list if there is an empty assignedTo string")
			@Test
			void shouldReturnUnfilteredListOnEmpty() {
				var emptyFilterCriteria = filterCriteriaBuilder.filterByAssignedTo(false).assignedTo(StringUtils.EMPTY).build();

				var page = repository.findAll(createRequestWithFilterBy(emptyFilterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(2);
			}

			@DisplayName("should return an unfiltered list if there is a null assignedTo")
			@Test
			void shouldReturnUnfilteredListOnNull() {
				var nullFilterCriteria = filterCriteriaBuilder.assignedTo(null).build();

				var page = repository.findAll(createRequestWithFilterBy(nullFilterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(2);
			}

			@Test
			void shouldReturnUnassignedVorgang() {
				var filterCriteria = filterCriteriaBuilder.assignedTo(StringUtils.EMPTY).build();

				var page = repository.findAll(createRequestWithFilterBy(filterCriteria));

				assertThat(page.getTotalElements()).isEqualTo(1);
			}
		}

		@DisplayName("Has next wiedervorlage frist")
		@Nested
		class TestHasNextWiedervorlageFrist {

			@Test
			void shouldHaveTotalElements() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
						.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(1, ChronoUnit.DAYS))).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id("2").clientAttributes(null).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());

				var vorgaenge = repository.findAll(createRequestWithFilterBy(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(true).build()));

				assertThat(vorgaenge.getTotalElements()).isEqualTo(1);
			}

			@Test
			void shouldReturnWhereNextWiedervorlageExistsOnly() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null)
						.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now())).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().clientAttributes(null).id(null).build());

				var vorgaenge = repository.findAll(createRequestWithFilterBy(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(true).build()));

				assertThat(vorgaenge.getContent()).hasSize(1);
			}

			@Nested
			class TestWithFindVorgangQuery {

				@Test
				void shouldReturnWhereNextWiedervorlageExistsOnly() {
					mongoOperations.save(VorgangTestFactory.createBuilder().id(null)
							.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now())).build());
					mongoOperations.save(VorgangTestFactory.createBuilder().clientAttributes(null).id(null).build());

					var vorgaenge = repository.findAll(createFindVorgangRequest());

					assertThat(vorgaenge.getContent()).hasSize(1);
				}

				private FindVorgangRequest createFindVorgangRequest() {
					return FindVorgangRequestTestFactory.createBuilder()
							.searchBy(StringUtils.EMPTY)
							.findVorgangQuery(FindVorgangQuery.builder()
									.logicalOperator(LogicalOperator.AND)
									.operator(OperatorBuilder
											.from(GrpcQueryOperator.EXISTS)
											.fieldPath("ClientAttribute.Alfa.nextWiedervorlageFrist")
											.build())
									.build())
							.build();
				}
			}
		}

		private Page<VorgangHeader> callRepositoryFindAll(FindVorgangRequest request) {
			return repository.findAll(request);
		}
	}

	private FindVorgangRequest createRequestWithFilterBy(FilterCriteria filterBy) {
		return FindVorgangRequestTestFactory.createBuilder().searchBy(StringUtils.EMPTY).filterBy(filterBy).build();
	}

	private ClientAttributesMap createClientAttributesMapWithWiedervorlageNextFrist(LocalDate nextFrist) {
		return ClientAttributesMapTestFactory.createWithAttribute(buildWiedervorlageNextFristClientAttribute(nextFrist));
	}

	private ClientAttribute buildWiedervorlageNextFristClientAttribute(LocalDate nextFrist) {
		return ClientAttributeTestFactory.createBuilder()
				.clientName(CriteriaUtil.DEFAULT_CLIENT)
				.attributeName(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST)
				.stringValue(Optional.of(nextFrist.toString())).build();
	}

	private Vorgang buildWithOrganisationseinheitId(String organisationseinheitenId) {
		var eingang = buildEingangWithOrganisationseinheitenId(organisationseinheitenId);
		return VorgangTestFactory.createBuilder().id(null).clearEingangs().eingang(eingang).build();
	}

	private Eingang buildEingangWithOrganisationseinheitenId(String organisationseinheitenId) {
		var zustaendigeStelle = ZustaendigeStelleTestFactory.createBuilder().organisationseinheitenId(organisationseinheitenId).build();
		return EingangTestFactory.createBuilder().zustaendigeStelle(zustaendigeStelle).build();
	}

	@Nested
	class TestWithClientAttribute {

		final FindVorgangRequest request = FindVorgangRequestTestFactory.createBuilder()
				.orderBy(OrderCriteria.PRIORITY)
				.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().build())
				.searchBy(StringUtils.EMPTY).build();

		@Test
		void shouldHaveClientAttribute() {
			persistAttributes(buildPersistableAttributes(ClientAttributeTestFactory.create()));

			var attributes = loadAttributes();

			assertThat(attributes).hasSize(1).containsKey(ClientAttributeTestFactory.CLIENT_NAME);
		}

		private ClientAttributesMap buildPersistableAttributes(ClientAttribute... attributes) {
			var attributeMap = new ClientAttributeMap();
			Arrays.stream(attributes).forEach(attrib -> attributeMap.put(attrib.getAttributeName(), attrib));
			ClientAttributesMap attributesMap = new ClientAttributesMap();
			attributesMap.put(attributes[0].getClientName(), attributeMap);

			return attributesMap;
		}

		private void persistAttributes(ClientAttributesMap attributesMap) {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).clientAttributes(attributesMap).build()).getId();
		}

		private ClientAttributesMap loadAttributes() {
			var vorgangList = repository.findAll(request).getContent();

			assertThat(vorgangList).hasSize(1);
			var attributes = vorgangList.get(0).getClientAttributes();
			return attributes;
		}
	}
}