/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;

import java.io.IOException;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.vorgang.vorgang.AntragstellerTestFactory;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

@ITCase
class ForwardingServiceITCase {

	@Autowired
	private ForwardingService service;

	@Test
	void getRedirectTemplate() {
		var template = service.getTemplate(ForwardingService.MAIL_TEMPLATE);

		assertThat(template).isNotNull();
	}

	@Test
	void buildLogoResource() throws IOException {
		var logoAttachment = service.buildImageAttachment(ForwardingService.PATH_LOGO_ITVSH, "id");

		assertThat(logoAttachment).isNotNull();
		assertThat(logoAttachment.getDataSource().getInputStream()).isNotNull().isNotEmpty();
	}

	@Nested
	class TestFillConfirmationTemplate {

		@Nested
		class Anrede {
			@Test
			void shouldBeGeneric() {
				var vorgang = createWithAnrede(null);

				var msg = service.fillConfirmationTemplate(vorgang);

				assertThat(msg).contains("Sehr geehrte Damen und Herren");
			}

			@Test
			void shouldBeHerr() {
				var vorgang = createWithAnrede("Herr");

				var msg = service.fillConfirmationTemplate(vorgang);

				assertThat(msg).contains("Sehr geehrter Herr " + AntragstellerTestFactory.NACHNAME);
			}

			private Vorgang createWithAnrede(String anrede) {
				return VorgangTestFactory.createBuilder().clearEingangs().eingang(
						EingangTestFactory.createBuilder().antragsteller(AntragstellerTestFactory.createBuilder().anrede(anrede).build()).build())
						.build();
			}
		}

		@Nested
		class ZustaendigeStelle {
			@Test
			void shouldBePresent() {
				var msg = service.fillConfirmationTemplate(createWithZustStelle("wichtiges Amt"));

				assertThat(msg).contains("Behörde wichtiges Amt ist");
			}

			@Test
			void shouldBeEmpty() {
				var msg = service.fillConfirmationTemplate(createWithZustStelle(null));

				assertThat(msg).contains("Behörde ist");
			}

			private Vorgang createWithZustStelle(String bezeichung) {
				return VorgangTestFactory.createBuilder().clearEingangs().eingang(
						EingangTestFactory.createBuilder()
								.zustaendigeStelle(ZustaendigeStelleTestFactory.createBuilder().bezeichnung(bezeichung).build())
								.build())
						.build();
			}
		}

		@Nested
		class TestDisclaimer {

			@Test
			void shouldBeAbsent() {
				ReflectionTestUtils.setField(service, "production", true);

				var msg = service.fillConfirmationTemplate(VorgangTestFactory.create());

				assertThat(msg).doesNotContain("TESTMAIL");
			}

			@Test
			void shouldBePresent() {
				ReflectionTestUtils.setField(service, "production", false);

				var msg = service.fillConfirmationTemplate(VorgangTestFactory.create());

				assertThat(msg).contains("TESTMAIL");
			}
		}
	}
}
