/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import static java.util.Objects.*;

import java.time.LocalDate;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.tuple.Pair;

import de.ozgcloud.vorgang.vorgang.GrpcFormField;

class FormDataMappingUtils {

	static final String VALUE_KEY = "value";
	static final String LABEL_KEY = "label";
	static final String CONTROL_DATA_KEY = "controlData";

	static final String META_DATA_KEY = "metadata";

	private FormDataMappingUtils() {
	}

	static GrpcFormField buildFormField(String name, String value) {
		return buildFormField(name, value, null);
	}

	static GrpcFormField buildFormField(String name, String value, String label) {
		var formFieldBuilder = GrpcFormField.newBuilder().setName(name).setValue(value);
		if (nonNull(label)) {
			formFieldBuilder.setLabel(label);
		}
		return formFieldBuilder.build();
	}

	static Map<String, Object> pairAccumulator(Map<String, Object> map, Pair<String, Map<String, Object>> pair) {
		return addToMap(map, pair.getLeft(), pair.getRight());
	}

	static Map<String, Object> combiner(Map<String, Object> map, Map<String, Object> toAddMap) {
		toAddMap.forEach((key, value) -> addToMap(map, key, value));
		return map;
	}

	static Map<String, Object> addToMap(Map<String, Object> map, String name, Object value) {
		map.compute(name, (k, v) -> isNull(v) ? value : merge(v, value));
		return map;
	}

	static Object merge(Object currentValue, Object newValue) {
		var result = buildListFrom(currentValue);
		result.addAll(buildListFrom(newValue));
		return Collections.unmodifiableList(result);
	}

	private static List<Object> buildListFrom(Object value) {
		if (value instanceof Collection<?> values) {
			return new LinkedList<>(values);
		} else {
			return new LinkedList<>(List.of(value));
		}
	}

	static boolean isSubForm(Object value) {
		return (!isValue(value) && getValueObject(value) instanceof Map);
	}

	static boolean isValue(Object value) {
		if (value instanceof Map<?, ?> valueMap) {
			return valueMap.containsKey(VALUE_KEY) &&
					(valueMap.get(VALUE_KEY) instanceof String
							|| valueMap.get(VALUE_KEY) instanceof Number
							|| valueMap.get(VALUE_KEY) instanceof LocalDate
							|| valueMap.get(VALUE_KEY) instanceof Boolean);
		} else {
			return value instanceof String;
		}
	}

	static Object getValueObject(Object valueContainer) {
		if (valueContainer instanceof Map<?, ?> valueMap) {
			if (valueMap.containsKey(VALUE_KEY)) {
				return valueMap.get(VALUE_KEY);
			}
			return valueContainer;
		} else {
			return valueContainer;
		}
	}

	static String getValue(Object valueContainer) {
		return getValueObject(valueContainer).toString();
	}

	@SuppressWarnings("unchecked")
	static Map<String, Object> getValueMap(Map<String, Object> valueContainer) {
		if (valueContainer.containsKey(VALUE_KEY)) {
			return (Map<String, Object>) valueContainer.get(VALUE_KEY);
		}
		return valueContainer;
	}

	static Optional<String> getLabel(Object valueContainer) {
		if (valueContainer instanceof Map<?, ?> valueMap) {
			return Optional.ofNullable((String) valueMap.get(LABEL_KEY));
		}
		return Optional.empty();
	}

	static boolean isMapWithLabel(Map<?, ?> map) {
		return nonNull(map) && map.size() == 2
				&& map.containsKey(LABEL_KEY) && map.containsKey(VALUE_KEY);
	}

	@SuppressWarnings("unchecked")
	public static <A, B> Map<A, B> getMap(Map<String, Object> map, String key, Map<A, B> defaultMap) {
		return MapUtils.getMap(map, key, defaultMap);
	}
}
